from infrahub import lock
from infrahub.core.branch import Branch
from infrahub.core.constants.infrahubkind import REPOSITORYVALIDATOR, USERVALIDATOR
from infrahub.core.convert_object_type.object_conversion import (
    ConversionFieldInput,
    convert_object_type,
    validate_conversion,
)
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreReadOnlyRepository, CoreRepository
from infrahub.core.schema import NodeSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.message_bus.messages import RefreshRegistryBranches
from infrahub.repositories.create_repository import RepositoryFinalizer
from infrahub.workers.dependencies import get_message_bus


async def convert_repository_type(
    repository: CoreRepository | CoreReadOnlyRepository,
    target_schema: NodeSchema,
    mapping: dict[str, ConversionFieldInput],
    branch: Branch,
    db: InfrahubDatabase,
    repository_post_creator: RepositoryFinalizer,
) -> Node:
    """Delete the node and return the new created one. If creation fails, the node is not deleted, and raise an error.
    An extra check is performed on input node peers relationships to make sure they are still valid."""

    repo_name = repository.name.value
    async with lock.registry.get(name=repo_name, namespace="repository"):
        async with db.start_transaction() as dbt:
            timestamp_before_conversion = Timestamp()

            # Fetch validators before deleting the repository otherwise validator-repository would no longer exist
            user_validators = await NodeManager.query(
                db=dbt, schema=USERVALIDATOR, prefetch_relationships=True, filters={"repository__id": repository.id}
            )
            repository_validators = await NodeManager.query(
                db=dbt,
                schema=REPOSITORYVALIDATOR,
                prefetch_relationships=True,
                filters={"repository__id": repository.id},
            )
            new_repository = await convert_object_type(
                node=repository,  # type: ignore[arg-type]
                target_schema=target_schema,
                mapping=mapping,
                branch=branch,
                db=dbt,
            )

            for user_validator in user_validators:
                await user_validator.repository.update(db=dbt, data=new_repository)
                await user_validator.repository.save(db=dbt)

            for repository_validator in repository_validators:
                await repository_validator.repository.update(db=dbt, data=new_repository)
                await repository_validator.repository.save(db=dbt)

            await validate_conversion(
                deleted_node=repository,  # type: ignore[arg-type]
                branch=branch,
                db=dbt,
                timestamp_before_conversion=timestamp_before_conversion,
            )

        # Refresh outside the transaction otherwise other workers would pull outdated branch objects.
        message_bus = await get_message_bus()
        await message_bus.send(RefreshRegistryBranches())

        # Following call involve a potential update of `commit` value of the newly created repository
        # that would be done from another database connection so it can't be performed within above transaction.
        # Also note since the conversion can only be performed on main branch here, it is fine that we do it
        # after having updating other branches status to NEEDS_REBASE.
        await repository_post_creator.post_create(
            branch=branch,
            obj=new_repository,  # type: ignore
            db=db,
            delete_on_connectivity_failure=False,
        )

        # Delete the RepositoryGroup associated with the old repository, as a new one was created for the new repository.
        repository_groups = (await repository.groups_objects.get_peers(db=db)).values()
        for repository_group in repository_groups:
            await NodeManager.delete(db=db, branch=branch, nodes=[repository_group], cascade_delete=False)

    return new_repository



from __future__ import annotations

from typing import Any

from graphql import (
    DocumentNode,
    FieldNode,
    GraphQLError,
    GraphQLSchema,
    OperationDefinitionNode,
    OperationType,
    parse,
    validate,
)
from pydantic import BaseModel

from .utils import calculate_dict_depth, calculate_dict_height, extract_fields


class GraphQLQueryVariable(BaseModel):
    name: str
    type: str
    required: bool = False
    default_value: Any | None = None


class GraphQLOperation(BaseModel):
    name: str | None = None
    operation_type: OperationType


class GraphQLQueryAnalyzer:
    def __init__(self, query: str, schema: GraphQLSchema | None = None) -> None:
        self.query: str = query
        self.schema: GraphQLSchema | None = schema
        self.document: DocumentNode = parse(self.query)
        self._fields: dict | None = None

    @property
    def is_valid(self) -> tuple[bool, list[GraphQLError] | None]:
        if self.schema is None:
            return False, [GraphQLError("Schema is not provided")]

        errors = validate(schema=self.schema, document_ast=self.document)
        if errors:
            return False, errors

        return True, None

    @property
    def nbr_queries(self) -> int:
        return len(self.document.definitions)

    @property
    def operations(self) -> list[GraphQLOperation]:
        operations = []
        for definition in self.document.definitions:
            if not isinstance(definition, OperationDefinitionNode):
                continue
            operation_type = definition.operation
            for field_node in definition.selection_set.selections:
                if not isinstance(field_node, FieldNode):
                    continue
                operations.append(GraphQLOperation(operation_type=operation_type, name=field_node.name.value))
        return operations

    @property
    def contains_mutation(self) -> bool:
        return any(op.operation_type == OperationType.MUTATION for op in self.operations)

    @property
    def variables(self) -> list[GraphQLQueryVariable]:
        response = []
        for definition in self.document.definitions:
            variable_definitions = getattr(definition, "variable_definitions", None)
            if not variable_definitions:
                continue
            for variable in variable_definitions:
                data = {"name": variable.variable.name.value}
                variable_type = variable.type
                non_null = variable_type.kind == "non_null_type"

                # This should not iterate a lot but it allows to inspect non-nullable iterables
                while hasattr(variable_type, "type"):
                    variable_type = variable_type.type
                data["type"] = variable_type.name.value

                if variable.default_value:
                    if data["type"] == "Int":
                        data["default_value"] = int(variable.default_value.value)
                    else:
                        data["default_value"] = variable.default_value.value

                if not data.get("default_value") and non_null:
                    data["required"] = True

                response.append(GraphQLQueryVariable(**data))

        return response

    async def calculate_depth(self) -> int:
        """Number of nested levels in the query"""
        fields = await self.get_fields()
        return calculate_dict_depth(data=fields)

    async def calculate_height(self) -> int:
        """Total number of fields requested in the query"""
        fields = await self.get_fields()
        return calculate_dict_height(data=fields)

    async def get_fields(self) -> dict[str, Any]:
        if not self._fields:
            fields = {}
            for definition in self.document.definitions:
                if not isinstance(definition, OperationDefinitionNode):
                    continue
                fields_to_update = await extract_fields(definition.selection_set)
                if fields_to_update is not None:
                    fields.update(fields_to_update)
            self._fields = fields
        return self._fields



from infrahub.core.diff.enricher.path_identifier import DiffPathIdentifierEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffPathIdentifierEnricherDependency(DependencyBuilder[DiffPathIdentifierEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffPathIdentifierEnricher:
        return DiffPathIdentifierEnricher(db=context.db)



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Any

from netaddr import IPSet

from infrahub import lock
from infrahub.core import registry
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.query.ipam import get_subnets
from infrahub.core.query.resource_manager import (
    PrefixPoolGetReserved,
    PrefixPoolSetReserved,
)
from infrahub.exceptions import ValidationError
from infrahub.pools.prefix import get_next_available_prefix

from .. import Node
from ..lock_utils import RESOURCE_POOL_LOCK_NAMESPACE

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.ipam.constants import IPNetworkType
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class CoreIPPrefixPool(Node):
    async def get_resource(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        identifier: str | None = None,
        data: dict[str, Any] | None = None,
        prefixlen: int | None = None,
        member_type: str | None = None,
        prefix_type: str | None = None,
        at: Timestamp | None = None,
    ) -> Node:
        async with lock.registry.get(name=self.get_id(), namespace=RESOURCE_POOL_LOCK_NAMESPACE):
            # Check if there is already a resource allocated with this identifier
            # if not, pull all existing prefixes and allocated the next available
            if identifier:
                query_get = await PrefixPoolGetReserved.init(db=db, pool_id=self.id, identifier=identifier)
                await query_get.execute(db=db)
                result = query_get.get_result()
                if result:
                    prefix = result.get_node("prefix")
                    # TODO add support for branch, if the node is reserved with this id in another branch we should return an error
                    node = await registry.manager.get_one(db=db, id=prefix.get("uuid"), branch=branch)
                    if node:
                        return node

            ip_namespace = await self.ip_namespace.get_peer(db=db)  # type: ignore[attr-defined]

            data = data or {}

            prefixlen = prefixlen or data.get("prefixlen", None) or self.default_prefix_length.value  # type: ignore[attr-defined]
            if not prefixlen:
                raise ValueError(
                    f"IPPrefixPool: {self.name.value} | "  # type: ignore[attr-defined]
                    "A prefixlen or a default_value must be provided to allocate a new prefix"
                )

            next_prefix = await self.get_next(db=db, prefixlen=prefixlen)

            prefix_type = prefix_type or data.get("prefix_type", None) or self.default_prefix_type.value  # type: ignore[attr-defined]
            if not prefix_type:
                raise ValueError(
                    f"IPPrefixPool: {self.name.value} | "  # type: ignore[attr-defined]
                    "A prefix_type or a default_value type must be provided to allocate a new prefix"
                )

            member_type = member_type or data.get("member_type", None) or self.default_member_type.value.value  # type: ignore[attr-defined]
            data["member_type"] = member_type

            target_schema = registry.get_node_schema(name=prefix_type, branch=branch)
            node = await Node.init(db=db, schema=target_schema, branch=branch, at=at)
            try:
                await node.new(db=db, prefix=str(next_prefix), ip_namespace=ip_namespace, **data)
            except ValidationError as exc:
                raise ValueError(f"IPPrefixPool: {self.name.value} | {exc!s}") from exc  # type: ignore[attr-defined]
            await node.save(db=db, at=at)
            reconciler = IpamReconciler(db=db, branch=branch)
            await reconciler.reconcile(ip_value=next_prefix, namespace=ip_namespace.id, node_uuid=node.get_id())

            if identifier:
                query_set = await PrefixPoolSetReserved.init(
                    db=db, pool_id=self.id, identifier=identifier, prefix_id=node.id, at=at
                )
                await query_set.execute(db=db)

            return node

    async def get_next(self, db: InfrahubDatabase, prefixlen: int) -> IPNetworkType:
        resources = await self.resources.get_peers(db=db)  # type: ignore[attr-defined]
        ip_namespace = await self.ip_namespace.get_peer(db=db)  # type: ignore[attr-defined]

        try:
            weighted_resources = sorted(resources.values(), key=lambda r: r.allocation_weight.value or 0, reverse=True)
        except AttributeError:
            weighted_resources = list(resources.values())

        for resource in weighted_resources:
            subnets = await get_subnets(
                db=db,
                ip_prefix=ipaddress.ip_network(resource.prefix.value),  # type: ignore[attr-defined]
                namespace=ip_namespace,
                branch=self._branch,
                branch_agnostic=True,
            )

            pool = IPSet([resource.prefix.value])
            for subnet in subnets:
                pool.remove(addr=str(subnet.prefix))

            try:
                prefix_ver = ipaddress.ip_network(resource.prefix.value).version
                next_available = get_next_available_prefix(pool=pool, prefix_length=prefixlen, prefix_ver=prefix_ver)
                return next_available
            except ValueError:
                continue

        raise IndexError("No more resources available")



from infrahub.utils import InfrahubStringEnum


class BranchStatus(InfrahubStringEnum):
    OPEN = "OPEN"
    NEED_REBASE = "NEED_REBASE"
    NEED_UPGRADE_REBASE = "NEED_UPGRADE_REBASE"
    DELETING = "DELETING"



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Literal

from netaddr import IPSet

if TYPE_CHECKING:
    from infrahub.core.ipam.constants import IPNetworkType


def get_next_available_prefix(
    pool: IPSet, prefix_length: int | None = None, prefix_ver: Literal[4, 6] = 4
) -> IPNetworkType:
    """Get the next available prefix of a given prefix length from an IPSet.

    Args:
        pool: netaddr IPSet object with available subnets
        prefix_length: length of the desired prefix
        prefix_ver: IPSet can contain a mix of IPv4 and IPv6 subnets. This parameter specifies the IP version of prefix to acquire.

    Raises:
        ValueError: If there are no available subnets in the pool
    """
    prefix_ver_map = {4: ipaddress.IPv4Network, 6: ipaddress.IPv6Network}

    filtered_pool = IPSet([])
    for subnet in pool.iter_cidrs():
        if isinstance(ipaddress.ip_network(str(subnet)), prefix_ver_map[prefix_ver]):
            filtered_pool.add(subnet)

    for cidr in filtered_pool.iter_cidrs():
        if prefix_length is None:
            return cidr

        if cidr.prefixlen <= prefix_length:
            next_available = ipaddress.ip_network(f"{cidr.network}/{prefix_length}")
            return next_available

    raise ValueError("No available subnets in pool")



from __future__ import annotations

import asyncio
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any

import typer
import yaml
from pydantic import ValidationError
from rich.console import Console

from ..async_typer import AsyncTyper
from ..ctl.client import initialize_client
from ..ctl.utils import catch_exception, init_logging
from ..queries import SCHEMA_HASH_SYNC_STATUS
from ..schema import SchemaWarning
from ..yaml import SchemaFile
from .parameters import CONFIG_PARAM
from .utils import load_yamlfile_from_disk_and_exit

if TYPE_CHECKING:
    from .. import InfrahubClient

app = AsyncTyper()
console = Console()


@app.callback()
def callback() -> None:
    """
    Manage the schema in a remote Infrahub instance.
    """


def validate_schema_content_and_exit(client: InfrahubClient, schemas: list[SchemaFile]) -> None:
    has_error: bool = False
    for schema_file in schemas:
        try:
            client.schema.validate(data=schema_file.payload)
        except ValidationError as exc:
            console.print(f"[red]Schema not valid, found '{len(exc.errors())}' error(s) in {schema_file.location}")
            has_error = True
            for error in exc.errors():
                loc_str = [str(item) for item in error["loc"]]
                console.print(f"  '{'/'.join(loc_str)}' | {error['msg']} ({error['type']})")

    if has_error:
        raise typer.Exit(1)


def display_schema_load_errors(response: dict[str, Any], schemas_data: list[SchemaFile]) -> None:
    console.print("[red]Unable to load the schema:")
    if "detail" not in response:
        handle_non_detail_errors(response=response)
        return

    for error in response["detail"]:
        loc_path = error.get("loc", [])
        if not valid_error_path(loc_path=loc_path):
            continue

        # if the len of the path is equal to 6, the error is at the root of the object
        # if the len of the path is higher than 6, the error is in an attribute or a relationships
        schema_index = int(loc_path[2])
        node_index = int(loc_path[4])
        node = get_node(schemas_data=schemas_data, schema_index=schema_index, node_index=node_index)

        if not node:
            console.print("Node data not found.")
            continue

        if len(loc_path) == 6:
            loc_type = loc_path[-1]
            input_str = error.get("input", None)
            error_message = f"{loc_type} ({input_str}) | {error['msg']} ({error['type']})"
            console.print(
                f"  Node: {node.get('namespace', None)}{node.get('name', None)} | {error_message}", markup=False
            )

        elif len(loc_path) > 6:
            loc_type = loc_path[5]
            error_data = node[loc_type]
            attribute = loc_path[6]

            if isinstance(attribute, str):
                input_label = None
                for data in error_data:
                    if data.get(attribute) is not None:
                        input_label = data.get("name", None)
                        break
            else:
                input_label = error_data[attribute].get("name", None)

            input_str = error.get("input", None)
            error_message = f"{loc_type[:-1].title()}: {input_label} ({input_str}) | {error['msg']} ({error['type']})"
            console.print(
                f"  Node: {node.get('namespace', None)}{node.get('name', None)} | {error_message}", markup=False
            )


def handle_non_detail_errors(response: dict[str, Any]) -> None:
    if "error" in response:
        console.print(f"  {response.get('error')}")
    elif "errors" in response:
        for error in response["errors"]:
            console.print(f"  {error.get('message')}")
    else:
        console.print(f"  '{response}'")


def valid_error_path(loc_path: list[Any]) -> bool:
    return len(loc_path) >= 6 and loc_path[0] == "body" and loc_path[1] == "schemas"


def get_node(schemas_data: list[SchemaFile], schema_index: int, node_index: int) -> dict | None:
    if schema_index < len(schemas_data) and node_index < len(schemas_data[schema_index].payload["nodes"]):
        return schemas_data[schema_index].payload["nodes"][node_index]
    return None


@app.command()
@catch_exception(console=console)
async def load(
    schemas: list[Path],
    debug: bool = False,
    branch: str = typer.Option(None, help="Branch on which to load the schema."),
    wait: int = typer.Option(0, help="Time in seconds to wait until the schema has converged across all workers"),
    _: str = CONFIG_PARAM,
) -> None:
    """Load one or multiple schema files into Infrahub."""

    init_logging(debug=debug)

    schemas_data = load_yamlfile_from_disk_and_exit(paths=schemas, file_type=SchemaFile, console=console)
    schema_definition = "schema" if len(schemas_data) == 1 else "schemas"
    client = initialize_client()
    validate_schema_content_and_exit(client=client, schemas=schemas_data)

    start_time = time.time()
    response = await client.schema.load(schemas=[item.payload for item in schemas_data], branch=branch)
    loading_time = time.time() - start_time

    if response.errors:
        display_schema_load_errors(response=response.errors, schemas_data=schemas_data)
        raise typer.Exit(1)

    if response.schema_updated:
        for schema_file in schemas_data:
            console.print(f"[green] schema '{schema_file.location}' loaded successfully")
    else:
        console.print("[green] The schema in Infrahub was already up to date, no changes were required")

    console.print(f"[green] {len(schemas_data)} {schema_definition} processed in {loading_time:.3f} seconds.")

    _display_schema_warnings(console=console, warnings=response.warnings)

    if response.schema_updated and wait:
        waited = 0
        continue_waiting = True
        while continue_waiting:
            status = await client.execute_graphql(query=SCHEMA_HASH_SYNC_STATUS, branch_name=branch)
            if status["InfrahubStatus"]["summary"]["schema_hash_synced"]:
                console.print("[green] Schema updated on all workers.")
                continue_waiting = False
            else:
                if waited >= wait:
                    console.print("[red] Schema is still not in sync after the specified waiting time")
                    raise typer.Exit(1)
                console.print("[yellow] Waiting for schema to sync across all workers")
                waited += 1
                await asyncio.sleep(delay=1)


@app.command()
@catch_exception(console=console)
async def check(
    schemas: list[Path],
    debug: bool = False,
    branch: str = typer.Option(None, help="Branch on which to check the schema."),
    _: str = CONFIG_PARAM,
) -> None:
    """Check if schema files are valid and what would be the impact of loading them with Infrahub."""

    init_logging(debug=debug)

    schemas_data = load_yamlfile_from_disk_and_exit(paths=schemas, file_type=SchemaFile, console=console)
    client = initialize_client()
    validate_schema_content_and_exit(client=client, schemas=schemas_data)

    success, response = await client.schema.check(schemas=[item.payload for item in schemas_data], branch=branch)

    if not success or not response:
        display_schema_load_errors(response=response or {}, schemas_data=schemas_data)
        return

    for schema_file in schemas_data:
        console.print(f"[green] schema '{schema_file.location}' is Valid!")

    warnings = response.pop("warnings", [])
    schema_warnings = [SchemaWarning.model_validate(warning) for warning in warnings]
    _display_schema_warnings(console=console, warnings=schema_warnings)
    if response == {"diff": {"added": {}, "changed": {}, "removed": {}}}:
        print("No diff")
    else:
        print(yaml.safe_dump(data=response, indent=4))


def _display_schema_warnings(console: Console, warnings: list[SchemaWarning]) -> None:
    for warning in warnings:
        console.print(
            f"[yellow] {warning.type.value}: {warning.message} [{', '.join([kind.display for kind in warning.kinds])}]"
        )



from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema

core_object_template = GenericSchema(
    name="ObjectTemplate",
    namespace="Core",
    include_in_menu=False,
    icon="mdi:pencil-ruler",
    description="Template to create pre-shaped objects.",
    label="Object Templates",
    display_labels=["template_name__value"],
    default_filter="template_name__value",
    uniqueness_constraints=[["template_name__value"]],
    attributes=[Attr(name="template_name", kind="Text", optional=False, unique=True, order_weight=1000)],
)

core_object_component_template = GenericSchema(
    name="ObjectComponentTemplate",
    namespace="Core",
    include_in_menu=False,
    icon="mdi:pencil-ruler",
    description="Component template to create pre-shaped objects.",
    label="Object Component Templates",
    display_labels=["template_name__value"],
    default_filter="template_name__value",
    attributes=[Attr(name="template_name", kind="Text", optional=False, order_weight=1000)],
)



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path


class ExporterInterface(ABC):
    @abstractmethod
    async def export(
        self, export_directory: Path, namespaces: list[str], branch: str, exclude: list[str] | None = None
    ) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from cachetools import TTLCache
from cachetools.keys import hashkey
from cachetools_async import cached
from git.exc import BadName, GitCommandError
from infrahub_sdk.exceptions import GraphQLError
from prefect import task
from prefect.cache_policies import NONE
from pydantic import Field

from infrahub.core.constants import InfrahubKind, RepositoryInternalStatus
from infrahub.exceptions import RepositoryError
from infrahub.git.integrator import InfrahubRepositoryIntegrator
from infrahub.log import get_logger

if TYPE_CHECKING:
    from infrahub_sdk.client import InfrahubClient

log = get_logger()


class InfrahubRepository(InfrahubRepositoryIntegrator):
    """
    Primary type of Git repository, with deep integration within Infrahub.

    Eventually we should rename this class InfrahubIntegratedRepository
    """

    @classmethod
    async def new(cls, update_commit_value: bool = True, **kwargs: Any) -> InfrahubRepository:
        self = cls(**kwargs)
        await self.create_locally(
            infrahub_branch_name=self.infrahub_branch_name, update_commit_value=update_commit_value
        )
        log.info("Created new repository locally.", repository=self.name)
        return self

    def get_commit_value(self, branch_name: str, remote: bool = False) -> str:
        branches = {}
        if remote:
            branches = self.get_branches_from_remote()
        else:
            branches = self.get_branches_from_local(include_worktree=False)

        if branch_name not in branches:
            raise ValueError(f"Branch {branch_name} not found.")

        return str(branches[branch_name].commit)

    async def create_branch_in_git(
        self, branch_name: str, branch_id: str | None = None, push_origin: bool = True
    ) -> bool:
        """Create new branch in the repository, assuming the branch has been created in the graph already."""

        response = await super().create_branch_in_git(branch_name=branch_name, branch_id=branch_id)
        if push_origin:
            await self.push(branch_name)

        return response

    async def sync(self, staging_branch: str | None = None) -> None:
        """Synchronize the repository with its remote origin and with the database.

        By default the sync will focus only on the branches pulled from origin that have some differences with the local one.
        """

        log.info("Starting the synchronization.", repository=self.name)

        await self.fetch()

        new_branches, updated_branches = await self.compare_local_remote()

        if not new_branches and not updated_branches:
            return

        log.debug(f"New Branches {new_branches}, Updated Branches {updated_branches}", repository=self.name)

        # TODO need to handle properly the situation when a branch is not valid.
        if self.internal_status == RepositoryInternalStatus.ACTIVE.value:
            for branch_name in new_branches:
                is_valid = self.validate_remote_branch(branch_name=branch_name)
                if not is_valid:
                    continue

                infrahub_branch = self._get_mapped_target_branch(branch_name=branch_name)
                try:
                    branch = await self.create_branch_in_graph(branch_name=infrahub_branch)
                except GraphQLError as exc:
                    if "already exist" not in exc.errors[0]["message"]:
                        raise
                    branch = await self.sdk.branch.get(branch_name=infrahub_branch)

                await self.create_branch_in_git(branch_name=branch.name, branch_id=branch.id, push_origin=True)

                commit = self.get_commit_value(branch_name=branch_name, remote=False)
                self.create_commit_worktree(commit=commit)
                await self.update_commit_value(branch_name=infrahub_branch, commit=commit)

                await self.import_objects_from_files(infrahub_branch_name=infrahub_branch, commit=commit)

            for branch_name in updated_branches:
                is_valid = self.validate_remote_branch(branch_name=branch_name)
                if not is_valid:
                    continue

                infrahub_branch = self._get_mapped_target_branch(branch_name=branch_name)

                commit_after = await self.pull(branch_name=branch_name)
                if isinstance(commit_after, str):
                    await self.import_objects_from_files(infrahub_branch_name=infrahub_branch, commit=commit_after)

                elif commit_after is True:
                    log.warning(
                        f"An update was detected but the commit remained the same after pull() ({commit_after}).",
                        repository=self.name,
                        branch=branch_name,
                    )

        await self._sync_staging(staging_branch=staging_branch, updated_branches=updated_branches)

    async def _sync_staging(self, staging_branch: str | None, updated_branches: list[str]) -> None:
        if (
            self.internal_status == RepositoryInternalStatus.STAGING.value
            and staging_branch
            and self.default_branch in updated_branches
        ):
            commit_after = await self.pull(branch_name=self.default_branch)
            if isinstance(commit_after, str):
                await self.import_objects_from_files(
                    git_branch_name=self.default_branch, infrahub_branch_name=staging_branch, commit=commit_after
                )

            elif commit_after is True:
                log.warning(
                    f"An update was detected but the commit remained the same after pull() ({commit_after}).",
                    repository=self.name,
                    branch=self.default_branch,
                )

    async def push(self, branch_name: str) -> bool:
        """Push a given branch to the remote Origin repository"""

        if not self.has_origin:
            return False

        log.debug(
            f"Pushing the latest update to the remote origin for the branch '{branch_name}'", repository=self.name
        )

        # TODO Catch potential exceptions coming from origin.push
        repo = self.get_git_repo_worktree(identifier=branch_name)
        remote_branch = self._get_mapped_remote_branch(branch_name=branch_name)
        repo.remotes.origin.push(remote_branch)

        return True

    async def merge(self, source_branch: str, dest_branch: str, push_remote: bool = True) -> bool:
        """Merge the source branch into the destination branch.

        After the rebase we need to resync the data
        """
        repo = self.get_git_repo_worktree(identifier=dest_branch)
        if not repo:
            raise ValueError(f"Unable to identify the worktree for the branch : {dest_branch}")

        commit_before = str(repo.head.commit)
        commit = self.get_commit_value(branch_name=source_branch, remote=False)

        try:
            repo.git.merge(commit)
        except GitCommandError as exc:
            repo.git.merge("--abort")
            raise RepositoryError(identifier=self.name, message=exc.stderr) from exc

        commit_after = str(repo.head.commit)

        if commit_after == commit_before:
            return False

        self.create_commit_worktree(commit_after)
        await self.update_commit_value(branch_name=dest_branch, commit=commit_after)
        if self.has_origin and push_remote:
            await self.push(branch_name=dest_branch)

        return str(commit_after)

    async def rebase(self, branch_name: str, source_branch: str = "main", push_remote: bool = True) -> bool:
        """Rebase the current branch with main.

        Technically we are not doing a Git rebase because it will change the git history
        We'll merge the content of the source_branch into branch_name instead to keep the history clear.

        TODO need to see how we manage conflict

        After the rebase we need to resync the data
        """

        response = await self.merge(dest_branch=branch_name, source_branch=source_branch, push_remote=push_remote)

        return response


class InfrahubReadOnlyRepository(InfrahubRepositoryIntegrator):
    """
    Repository with only read-only access to the remote repo
    """

    is_read_only: bool = True
    ref: str | None = Field(None, description="Ref to track on the external repository")

    @classmethod
    async def new(cls, **kwargs: Any) -> InfrahubReadOnlyRepository:
        if "ref" not in kwargs or "infrahub_branch_name" not in kwargs:
            raise ValueError("ref and infrahub_branch_name are mandatory to initialize a new Read-Only repository")

        self = cls(**kwargs)
        await self.create_locally(checkout_ref=self.ref, infrahub_branch_name=self.infrahub_branch_name)
        log.info("Created new repository locally.", repository=self.name)
        return self

    def get_commit_value(self, branch_name: str, remote: bool = False) -> str:  # noqa: ARG002
        """Always get the latest commit for this repository's ref on the remote"""
        git_repo = self.get_git_repo_main()
        git_repo.remotes.origin.fetch()

        refs = (f"origin/{self.ref}", self.ref)
        commit = None
        for possible_ref in refs:
            try:
                commit = git_repo.commit(possible_ref)
                break
            except BadName:
                ...
        if not commit:
            log.error(f"No object found for refs {refs} on repository {self.name}")
            raise ValueError(f"Ref {self.ref} not found.")

        return str(commit)

    async def sync_from_remote(self, commit: str | None = None) -> None:
        if not commit:
            commit = self.get_commit_value(branch_name=self.ref, remote=True)
        local_branches = self.get_branches_from_local()
        if self.ref in local_branches and commit == local_branches[self.ref].commit:
            return
        self.create_commit_worktree(commit=commit)
        await self.import_objects_from_files(infrahub_branch_name=self.infrahub_branch_name, commit=commit)
        await self.update_commit_value(branch_name=self.infrahub_branch_name, commit=commit)


@cached(
    TTLCache(maxsize=100, ttl=30),
    key=lambda *_, **kwargs: hashkey(
        kwargs.get("repository_id"), kwargs.get("name"), kwargs.get("repository_kind"), kwargs.get("commit")
    ),
)
async def _get_initialized_repo(
    client: InfrahubClient, repository_id: str, name: str, repository_kind: str, commit: str | None = None
) -> InfrahubReadOnlyRepository | InfrahubRepository:
    if repository_kind == InfrahubKind.REPOSITORY:
        return await InfrahubRepository.init(id=repository_id, name=name, commit=commit, client=client)

    if repository_kind == InfrahubKind.READONLYREPOSITORY:
        return await InfrahubReadOnlyRepository.init(id=repository_id, name=name, commit=commit, client=client)

    raise NotImplementedError(f"The repository kind {repository_kind} has not been implemented")


@task(
    name="Fetch repository commit",
    description="Retrieve a git repository at a given commit, if it does not already exist locally",
    cache_policy=NONE,
)
async def get_initialized_repo(
    client: InfrahubClient, repository_id: str, name: str, repository_kind: str, commit: str | None = None
) -> InfrahubReadOnlyRepository | InfrahubRepository:
    return await _get_initialized_repo(
        client=client, repository_id=repository_id, name=name, repository_kind=repository_kind, commit=commit
    )



from __future__ import annotations

import copy
import logging
import re
from typing import Any

from ...exceptions import ValidationError
from ..range_expansion import MATCH_PATTERN, range_expansion
from .data_processor import DataProcessor

log = logging.getLogger("infrahub_sdk")


class RangeExpandDataProcessor(DataProcessor):
    """Process data with range expansion"""

    @classmethod
    async def process_data(
        cls,
        data: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Expand any item in data with range pattern in any value. Supports multiple fields, requires equal expansion length."""
        range_pattern = re.compile(MATCH_PATTERN)
        expanded = []
        for item in data:
            # Find all fields to expand
            expand_fields = {}
            for key, value in item.items():
                if isinstance(value, str) and range_pattern.search(value):
                    try:
                        expand_fields[key] = range_expansion(value)
                    except (ValueError, TypeError, KeyError):
                        # If expansion fails, treat as no expansion
                        log.debug(
                            f"Range expansion failed for value '{value}' in key '{key}'. Treating as no expansion."
                        )
                        expand_fields[key] = [value]
            if not expand_fields:
                expanded.append(item)
                continue
            # Check all expanded lists have the same length
            lengths = [len(v) for v in expand_fields.values()]
            if len(set(lengths)) > 1:
                raise ValidationError(
                    identifier="range_expansion",
                    message=f"Range expansion mismatch: fields expanded to different lengths: {lengths}",
                )
            n = lengths[0]
            # Zip expanded values and produce new items
            for i in range(n):
                new_item = copy.deepcopy(item)
                for key, values in expand_fields.items():
                    new_item[key] = values[i]
                expanded.append(new_item)
        return expanded



from collections import defaultdict
from typing import TYPE_CHECKING, Any

from infrahub_sdk import InfrahubClient
from infrahub_sdk.node import RelationshipManager
from infrahub_sdk.protocols import CoreArtifactDefinition, CoreCheckDefinition, CoreGroup
from infrahub_sdk.types import Order

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.database import InfrahubDatabase
from infrahub.generators.models import ProposedChangeGeneratorDefinition

from .models import RepositoryBranchInfo, RepositoryData

if TYPE_CHECKING:
    from infrahub.core.protocols import CoreGenericRepository


async def get_repositories_commit_per_branch(
    db: InfrahubDatabase,
) -> dict[str, RepositoryData]:
    """Get a list of all repositories and their commit on each branches.

    This method is similar to 'get_list_repositories' method in the Python SDK.

    NOTE: At some point, we should refactor this function to use a single Database query instead of one per branch
    """

    repositories: dict[str, RepositoryData] = {}

    for branch in list(registry.branch.values()):
        repos: list[CoreGenericRepository] = await NodeManager.query(
            db=db,
            branch=branch,
            fields={"id": None, "name": None, "commit": None, "internal_status": None},
            schema=InfrahubKind.GENERICREPOSITORY,
        )

        for repository in repos:
            repo_name = repository.name.value
            if repo_name not in repositories:
                repositories[repo_name] = RepositoryData(
                    repository_id=repository.get_id(),
                    repository_name=repo_name,
                    branches={},
                )

            repositories[repo_name].branches[branch.name] = repository.commit.value  # type: ignore[attr-defined]
            repositories[repo_name].branch_info[branch.name] = RepositoryBranchInfo(
                internal_status=repository.internal_status.value
            )

    return repositories


def _collect_parameter_first_segments(params: Any) -> set[str]:
    segments: set[str] = set()

    def _walk(value: Any) -> None:
        if isinstance(value, str):
            segment = value.split("__", 1)[0]
            if segment:
                segments.add(segment)
        elif isinstance(value, dict):
            for nested in value.values():
                _walk(nested)
        elif isinstance(value, (list, tuple, set)):
            for nested in value:
                _walk(nested)

    _walk(params)
    return segments


async def _prefetch_group_member_nodes(
    client: InfrahubClient,
    members: RelationshipManager,
    branch: str,
    required_fields: set[str],
) -> None:
    ids_per_kind: dict[str, set[str]] = defaultdict(set)
    for peer in members.peers:
        if peer.id and peer.typename:
            ids_per_kind[peer.typename].add(peer.id)

    if not ids_per_kind:
        return

    batch = await client.create_batch()

    for kind, ids in ids_per_kind.items():
        schema = await client.schema.get(kind=kind, branch=branch)

        # FIXME: https://github.com/opsmill/infrahub-sdk-python/pull/205
        valid_fields = set(schema.attribute_names) | set(schema.relationship_names)
        keep_relationships = set(schema.relationship_names) & required_fields
        cleaned_fields = valid_fields - required_fields

        kwargs: dict[str, Any] = {
            "kind": kind,
            "ids": list(ids),
            "branch": branch,
            "exclude": list(cleaned_fields),
            "populate_store": True,
            "order": Order(disable=True),
        }

        if keep_relationships:
            kwargs["include"] = list(keep_relationships)

        batch.add(task=client.filters, **kwargs)

    async for _ in batch.execute():
        pass


async def _fetch_definition_targets(
    client: InfrahubClient,
    branch: str,
    group_id: str,
    parameters: Any,
) -> CoreGroup:
    group = await client.get(
        kind=CoreGroup,
        id=group_id,
        branch=branch,
        include=["members"],
    )

    parameter_fields = _collect_parameter_first_segments(parameters)
    await _prefetch_group_member_nodes(
        client=client,
        members=group.members,
        branch=branch,
        required_fields=parameter_fields,
    )

    return group


async def fetch_artifact_definition_targets(
    client: InfrahubClient,
    branch: str,
    definition: CoreArtifactDefinition,
) -> CoreGroup:
    return await _fetch_definition_targets(
        client=client, branch=branch, group_id=definition.targets.id, parameters=definition.parameters.value
    )


async def fetch_check_definition_targets(
    client: InfrahubClient,
    branch: str,
    definition: CoreCheckDefinition,
) -> CoreGroup:
    return await _fetch_definition_targets(
        client=client, branch=branch, group_id=definition.targets.id, parameters=definition.parameters.value
    )


async def fetch_proposed_change_generator_definition_targets(
    client: InfrahubClient,
    branch: str,
    definition: ProposedChangeGeneratorDefinition,
) -> CoreGroup:
    return await _fetch_definition_targets(
        client=client, branch=branch, group_id=definition.group_id, parameters=definition.parameters
    )



from collections import defaultdict

from infrahub.core import registry
from infrahub.core.constants import RelationshipHierarchyDirection, RelationshipKind
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.query.node import NodeGetHierarchyQuery
from infrahub.core.query.relationship import RelationshipGetPeerQuery, RelationshipPeerData
from infrahub.core.schema import ProfileSchema, TemplateSchema
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import SchemaNotFoundError
from infrahub.log import get_logger

from ..model.path import (
    CalculatedDiffs,
    EnrichedDiffRoot,
    NodeIdentifier,
)
from ..parent_node_adder import DiffParentNodeAdder, ParentNodeAddRequest
from .interface import DiffEnricherInterface

log = get_logger()


class DiffHierarchyEnricher(DiffEnricherInterface):
    """Add hierarchy and parent/component nodes to diff even if the higher-level nodes are unchanged"""

    def __init__(self, db: InfrahubDatabase, parent_adder: DiffParentNodeAdder):
        self.db = db
        self.parent_adder = parent_adder

    async def enrich(
        self,
        enriched_diff_root: EnrichedDiffRoot,
        calculated_diffs: CalculatedDiffs | None = None,  # noqa: ARG002
    ) -> None:
        # A hierarchy can be defined in 2 ways
        # - A node has a relationship of kind parent
        # - A node is part of a hierarchy

        log.info("Beginning hierarchical diff enrichment...")
        self.parent_adder.initialize(enriched_diff_root=enriched_diff_root)
        node_rel_parent_map: dict[str, list[NodeIdentifier]] = defaultdict(list)
        node_hierarchy_map: dict[str, list[NodeIdentifier]] = defaultdict(list)

        for node in enriched_diff_root.nodes:
            try:
                schema_node = self.db.schema.get(
                    name=node.kind, branch=enriched_diff_root.diff_branch_name, duplicate=False
                )
            except SchemaNotFoundError:
                continue

            if isinstance(schema_node, ProfileSchema | TemplateSchema):
                continue

            if schema_node.has_parent_relationship:
                node_rel_parent_map[node.kind].append(node.identifier)
                continue

            try:
                hierarchy_schema = schema_node.get_hierarchy_schema(
                    db=self.db, branch=enriched_diff_root.diff_branch_name
                )
                node_hierarchy_map[hierarchy_schema.kind].append(node.identifier)
            except ValueError:
                pass

        await self._enrich_nodes_with_parent(enriched_diff_root=enriched_diff_root, node_map=node_rel_parent_map)
        await self._enrich_hierarchical_nodes(enriched_diff_root=enriched_diff_root, node_map=node_hierarchy_map)
        log.info("Hierarchical diff enrichment complete.")

    async def _enrich_hierarchical_nodes(
        self,
        enriched_diff_root: EnrichedDiffRoot,
        node_map: dict[str, list[NodeIdentifier]],
    ) -> None:
        diff_branch = registry.get_branch_from_registry(branch=enriched_diff_root.diff_branch_name)

        # Retrieve the ID of all ancestors
        for kind, node_identifiers in node_map.items():
            log.info(f"Beginning hierarchy enrichment for {kind} node, num_nodes={len(node_identifiers)}...")
            hierarchy_schema = self.db.schema.get(
                name=kind, branch=enriched_diff_root.diff_branch_name, duplicate=False
            )
            for node_identifier in node_identifiers:
                query = await NodeGetHierarchyQuery.init(
                    db=self.db,
                    direction=RelationshipHierarchyDirection.ANCESTORS,
                    node_id=node_identifier.uuid,
                    node_schema=hierarchy_schema,
                    branch=diff_branch,
                    hierarchical_ordering=True,
                )
                await query.execute(db=self.db)

                ancestors = list(query.get_relatives())

                if not ancestors:
                    continue

                node = enriched_diff_root.get_node(node_identifier=node_identifier)
                parent_rel = hierarchy_schema.get_relationship(name="parent")

                current_node = node
                for ancestor in ancestors:
                    ancestor_identifier = NodeIdentifier(uuid=ancestor.uuid, kind=ancestor.kind, db_id=ancestor.db_id)
                    parent_request = ParentNodeAddRequest(
                        node_identifier=current_node.identifier,
                        parent_identifier=ancestor_identifier,
                        parent_label="",
                        parent_rel_name=parent_rel.name,
                        parent_rel_identifier=parent_rel.get_identifier(),
                        parent_rel_cardinality=parent_rel.cardinality,
                        parent_rel_label=parent_rel.label or "",
                    )
                    parent = self.parent_adder.add_parent(parent_request=parent_request)

                    current_node = parent

    async def _enrich_nodes_with_parent(
        self, enriched_diff_root: EnrichedDiffRoot, node_map: dict[str, list[NodeIdentifier]]
    ) -> None:
        diff_branch = registry.get_branch_from_registry(branch=enriched_diff_root.diff_branch_name)

        parent_peers: dict[NodeIdentifier, RelationshipPeerData] = {}

        # Prepare a map to capture all parents that also have a parent
        node_parent_with_parent_map: dict[str, list[NodeIdentifier]] = defaultdict(list)

        # TODO Not gonna implement it now but technically we could check the content of the node to see if the parent relationship is present

        # Query the UUID of the parent
        for kind, node_identifiers in node_map.items():
            log.info(f"Beginning parent enrichment for {kind} node, num_nodes={len(node_identifiers)}...")
            schema_node = self.db.schema.get(name=kind, branch=enriched_diff_root.diff_branch_name, duplicate=False)

            parent_rel = [rel for rel in schema_node.relationships if rel.kind == RelationshipKind.PARENT][0]
            parent_schema = self.db.schema.get(
                name=parent_rel.peer, branch=enriched_diff_root.diff_branch_name, duplicate=False
            )

            query = await RelationshipGetPeerQuery.init(
                db=self.db,
                branch=diff_branch,
                source_ids=[ni.uuid for ni in node_identifiers],
                rel_type=DatabaseEdgeType.IS_RELATED.value,
                schema=parent_rel,
            )
            await query.execute(db=self.db)

            for peer in query.get_peers():
                source_identifier = NodeIdentifier(
                    uuid=str(peer.source_id), kind=peer.source_kind, db_id=peer.source_db_id
                )
                parent_peers[source_identifier] = peer
                if parent_schema.has_parent_relationship:
                    peer_identifier = NodeIdentifier(uuid=str(peer.peer_id), kind=peer.peer_kind, db_id=peer.peer_db_id)
                    node_parent_with_parent_map[parent_schema.kind].append(peer_identifier)

        # Check if the parent are already present
        # If parent is already in the list of node we need to add a relationship
        # If parent is not in the list of node, we need to add it
        diff_node_map = enriched_diff_root.get_node_map(
            node_uuids={source_identifier.uuid for source_identifier in parent_peers.keys()}
        )
        for node_identifier, peer_parent in parent_peers.items():
            # TODO check if we can optimize this part to avoid querying this multiple times
            node = diff_node_map[node_identifier]
            schema_node = self.db.schema.get(
                name=node.kind, branch=enriched_diff_root.diff_branch_name, duplicate=False
            )
            parent_rel = [rel for rel in schema_node.relationships if rel.kind == RelationshipKind.PARENT][0]

            peer_identifier = NodeIdentifier(
                uuid=str(peer_parent.peer_id), kind=peer_parent.peer_kind, db_id=peer_parent.peer_db_id
            )
            parent_request = ParentNodeAddRequest(
                node_identifier=node.identifier,
                parent_identifier=peer_identifier,
                parent_label="",
                parent_rel_name=parent_rel.name,
                parent_rel_identifier=parent_rel.get_identifier(),
                parent_rel_cardinality=parent_rel.cardinality,
                parent_rel_label=parent_rel.label or "",
            )
            self.parent_adder.add_parent(parent_request=parent_request)

        if node_parent_with_parent_map:
            await self._enrich_nodes_with_parent(
                enriched_diff_root=enriched_diff_root, node_map=node_parent_with_parent_map
            )



from infrahub.database import InfrahubDatabase


async def verify_no_duplicate_relationships(db: InfrahubDatabase) -> None:
    """
    Verify that no duplicate active relationships exist at the database level
    A duplicate is defined as
    - connecting the same two nodes
    - having the same identifier
    - having the same direction (inbound, outbound, bidirectional)
    - having the same branch
    A more thorough check that no duplicates exist at any point in time is possible, but more complex
    """
    query = """
MATCH (a:Node)-[e1:IS_RELATED {status: "active"}]-(rel:Relationship)-[e2:IS_RELATED {branch: e1.branch, status: "active"}]-(b:Node)
WHERE a.uuid <> b.uuid
AND e1.to IS NULL
AND e2.to IS NULL
WITH a, rel.name AS rel_name, b, e1.branch AS branch, CASE
    WHEN startNode(e1) = a AND startNode(e2) = rel THEN "out"
    WHEN startNode(e1) = rel AND startNode(e2) = b THEN "in"
    ELSE "bidir"
END AS direction, COUNT(*) AS num_duplicates
WHERE num_duplicates > 1
RETURN a.uuid AS node_id1, b.uuid AS node_id2, rel_name, branch, direction, num_duplicates
    """
    results = await db.execute_query(query=query)
    for result in results:
        node_id1 = result.get("node_id1")
        node_id2 = result.get("node_id2")
        rel_name = result.get("rel_name")
        branch = result.get("branch")
        direction = result.get("direction")
        num_duplicates = result.get("num_duplicates")
        raise ValueError(
            f"{num_duplicates} duplicate relationships ({branch=},{direction=}) between nodes '{node_id1}' and '{node_id2}'"
            f" with relationship name '{rel_name}'"
        )


async def verify_no_edges_added_after_node_delete(db: InfrahubDatabase) -> None:
    """
    Verify that no edges are added to a Node after it is deleted on a given branch
    """
    query = """
// ------------
// find deleted nodes
// ------------
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE e.status = "deleted" OR e.to IS NOT NULL
WITH DISTINCT n, e.branch AS delete_branch, e.branch_level AS delete_branch_level, CASE
    WHEN e.status = "deleted" THEN e.from
    ELSE e.to
END AS delete_time
// ------------
// find the edges added to the deleted node after the delete time
// ------------
MATCH (n)-[added_e]-(peer)
WHERE added_e.from > delete_time
AND type(added_e) <> "IS_PART_OF"
// if the node was deleted on a branch (delete_branch_level > 1), and then updated on main/global (added_e.branch_level = 1), we can ignore it
AND added_e.branch_level >= delete_branch_level
AND (added_e.branch = delete_branch OR delete_branch_level = 1)
WITH DISTINCT n, delete_branch, delete_time, added_e, peer AS added_peer
// ------------
// get the branched_from for the branch on which the node was deleted
// ------------
CALL (added_e) {
    MATCH (b:Branch {name: added_e.branch})
    RETURN b.branched_from AS added_e_branched_from
}
// ------------
// account for the following situations, given that the edge update time is after the node delete time
//  - deleted on main/global, updated on branch
//    - illegal if the delete is before branch.branched_from
//  - deleted on branch, updated on branch
//    - illegal
// ------------
WITH n, delete_branch, delete_time, added_e, added_peer
WHERE delete_branch = added_e.branch
OR delete_time < added_e_branched_from
RETURN n.uuid AS n_uuid, delete_branch, delete_time, added_e, added_peer
    """
    results = await db.execute_query(query=query)
    error_messages = []
    for result in results:
        n_uuid = result.get("n_uuid")
        delete_branch = result.get("delete_branch")
        delete_time = result.get("delete_time")
        added_e = result.get("added_e")
        added_e_branch = added_e.get("branch")
        added_e_from = added_e.get("from")
        added_peer = result.get("added_peer")
        message = (
            f"Node {n_uuid} was deleted on {delete_branch} at {delete_time} but has an {added_e.type} edge added on"
            f" branch {added_e_branch} at {added_e_from} to {added_peer.element_id}"
        )
        error_messages.append(message)
    if error_messages:
        raise ValueError(error_messages)



from infrahub.core.relationship.constraints.peer_relatives import RelationshipPeerRelativesConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class RelationshipPeerRelativesConstraintDependency(DependencyBuilder[RelationshipPeerRelativesConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipPeerRelativesConstraint:
        return RelationshipPeerRelativesConstraint(db=context.db, branch=context.branch)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.registry import registry
from infrahub.core.schema import GenericSchema
from infrahub.permissions import report_schema_permissions

if TYPE_CHECKING:
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.graphql.initialization import GraphqlContext


async def get_permissions(schema: MainSchemaTypes, graphql_context: GraphqlContext) -> dict[str, Any]:
    schema_objects = [schema]
    if isinstance(schema, GenericSchema):
        for node_name in schema.used_by:
            schema_objects.append(registry.schema.get(name=node_name, branch=graphql_context.branch, duplicate=False))

    response: dict[str, Any] = {"count": len(schema_objects), "edges": []}

    nodes = await report_schema_permissions(
        branch=graphql_context.branch, permission_manager=graphql_context.active_permissions, schemas=schema_objects
    )
    response["edges"] = [{"node": node} for node in nodes]

    return response



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class DeleteDuplicatedAttributesQuery(Query):
    name: str = "delete_duplicated_attributes"
    type: QueryType = QueryType.WRITE
    insert_return: bool = False
    insert_limit: bool = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// -------------
// get all the Nodes linked to multiple Attributes with the same name to drastically reduce the search space
// -------------
MATCH (n:Node)-[:HAS_ATTRIBUTE]->(attr:Attribute)
WITH DISTINCT n, attr
WITH n, attr.name AS attr_name, count(*) AS num_attrs
WHERE num_attrs > 1
// -------------
// for each Node-attr_name pair, get the possible duplicate Attributes
// -------------
MATCH (n)-[:HAS_ATTRIBUTE]->(dup_attr:Attribute {name: attr_name})
WITH DISTINCT n, dup_attr
// -------------
// get the branch(es) for each possible duplicate Attribute
// -------------
CALL (n, dup_attr) {
    MATCH (n)-[r:HAS_ATTRIBUTE {status: "active"}]->(dup_attr)
    WHERE r.to IS NULL
    AND NOT exists((n)-[:HAS_ATTRIBUTE {status: "deleted", branch: r.branch}]->(dup_attr))
    RETURN r.branch AS branch
}
// -------------
// get the latest update time for each duplicate Attribute on each branch
// -------------
CALL (dup_attr, branch) {
    MATCH (dup_attr)-[r {branch: branch}]-()
    RETURN max(r.from) AS latest_update
}
// -------------
// order the duplicate Attributes by latest update time
// -------------
WITH n, dup_attr, branch, latest_update
ORDER BY n, branch, dup_attr.name, latest_update DESC
// -------------
// for any Node-dup_attr_name pairs with multiple duplicate Attributes, keep the Attribute with the latest update
// on this branch and delete all the other edges on this branch for this Attribute
// -------------
WITH n, branch, dup_attr.name AS dup_attr_name, collect(dup_attr) AS dup_attrs_reverse_chronological
WHERE size(dup_attrs_reverse_chronological) > 1
WITH branch, tail(dup_attrs_reverse_chronological) AS dup_attrs_to_delete
UNWIND dup_attrs_to_delete AS dup_attr_to_delete
MATCH (dup_attr_to_delete)-[r {branch: branch}]-()
DELETE r
// -------------
// delete any orphaned Attributes
// -------------
WITH DISTINCT dup_attr_to_delete
WHERE NOT exists((dup_attr_to_delete)--())
DELETE dup_attr_to_delete
        """
        self.add_to_query(query)


class Migration040(GraphMigration):
    name: str = "040_duplicated_attributes"
    queries: Sequence[type[Query]] = [DeleteDuplicatedAttributesQuery]
    minimum_version: int = 39

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()



from rich.syntax import Syntax
from rich.traceback import Frame, Traceback


class Error(Exception):
    """pytest-infrahub Base exception."""


class InvalidResourceConfigError(Error):
    def __init__(self, resource_name: str) -> None:
        super().__init__(f"Improperly configured resource with name '{resource_name}'.")


class DirectoryNotFoundError(Error):
    def __init__(self, name: str, message: str = "") -> None:
        self.message = message or f"Unable to find directory {name!r}."
        super().__init__(self.message)


class FileNotValidError(Error):
    def __init__(self, name: str, message: str = "") -> None:
        self.message = message or f"Unable to access file {name!r}."
        super().__init__(self.message)


class OutputMatchError(Error):
    def __init__(self, name: str, message: str = "", differences: str = "") -> None:
        self.message = message or f"Rendered output does not match expected output for {name!r}."
        self.differences = differences
        super().__init__(self.message)


class Jinja2TransformError(Error):
    def __init__(self, name: str, message: str = "") -> None:
        self.message = message or f"Unexpected error happened while processing {name!r}."
        super().__init__(self.message)


class Jinja2TransformUndefinedError(Error):
    def __init__(self, name: str, rtb: Traceback, errors: list[tuple[Frame, Syntax]], message: str = "") -> None:
        self.rtb = rtb
        self.errors = errors
        self.message = message or f"Unable to render Jinja2 transform {name!r}."
        super().__init__(self.message)


class CheckDefinitionError(Error):
    def __init__(self, name: str, message: str = "") -> None:
        self.message = message or f"Check {name!r} is not properly defined."
        super().__init__(self.message)


class CheckResultError(Error):
    def __init__(self, name: str, message: str = "") -> None:
        self.message = message or f"Unexpected result for check {name!r}."
        super().__init__(self.message)


class PythonTransformDefinitionError(Error):
    def __init__(self, name: str, message: str = "") -> None:
        self.message = message or f"Python transform {name!r} is not properly defined."
        super().__init__(self.message)



from dataclasses import dataclass

from infrahub.core.constants import InfrahubKind
from infrahub.core.node import Node
from infrahub.core.query.ipam import IPPrefixUtilization
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase

from .constants import PrefixMemberType
from .size import get_prefix_space


@dataclass
class PrefixChildDetails:
    child_type: PrefixMemberType
    prefixlen: int
    ip_value: str


class PrefixUtilizationGetter:
    def __init__(self, db: InfrahubDatabase, ip_prefixes: list[Node], at: Timestamp | str | None = None) -> None:
        self.db = db
        self.ip_prefixes = ip_prefixes
        self.at = at
        self._has_data = False
        self._results_by_prefix_id: dict[str, dict[str, list[PrefixChildDetails]]] = {}

    async def _fetch_data(self) -> None:
        if self._has_data is False:
            await self._run_and_parse_query()
        self._has_data = True

    async def _run_and_parse_query(self) -> None:
        self._results_by_prefix_id = {}
        query = await IPPrefixUtilization.init(
            db=self.db,
            at=self.at,
            ip_prefixes=self.ip_prefixes,
            allocated_kinds=[InfrahubKind.IPPREFIX, InfrahubKind.IPADDRESS],
        )
        await query.execute(db=self.db)

        for result in query.get_results():
            prefix_node = result.get_node("pfx")
            prefix_id = str(prefix_node.get("uuid"))
            branch_name = str(result.get("branch"))
            child_node = result.get_node("child")
            if InfrahubKind.IPADDRESS in child_node.labels:
                child_type = PrefixMemberType.ADDRESS
            else:
                child_type = PrefixMemberType.PREFIX
            child_value_node = result.get_node("av")
            child_prefixlen = child_value_node.get("prefixlen")
            child_ip_value = child_value_node.get("value")

            if prefix_id not in self._results_by_prefix_id:
                self._results_by_prefix_id[prefix_id] = {}
            if branch_name not in self._results_by_prefix_id[prefix_id]:
                self._results_by_prefix_id[prefix_id][branch_name] = []
            self._results_by_prefix_id[prefix_id][branch_name].append(
                PrefixChildDetails(child_type=child_type, prefixlen=child_prefixlen, ip_value=child_ip_value)
            )

    async def get_children(
        self,
        ip_prefixes: list[Node] | None = None,
        prefix_member_type: PrefixMemberType | None = None,
        branch_names: list[str] | None = None,
    ) -> list[PrefixChildDetails]:
        await self._fetch_data()
        prefix_child_details_list: list[PrefixChildDetails] = []
        if ip_prefixes is None:
            ip_prefixes = self.ip_prefixes
        prefix_ids = {prefix.get_id() for prefix in ip_prefixes}
        prefix_ids &= set(self._results_by_prefix_id.keys())
        if not prefix_ids:
            return prefix_child_details_list
        for prefix_id in prefix_ids:
            child_details_by_branch = self._results_by_prefix_id[prefix_id]
            branch_names_to_check = branch_names or list(child_details_by_branch.keys())
            for branch_name in branch_names_to_check:
                for child_details in child_details_by_branch.get(branch_name, []):
                    if prefix_member_type and child_details.child_type != prefix_member_type:
                        continue
                    prefix_child_details_list.append(child_details)
        return prefix_child_details_list

    async def get_num_children_in_use(
        self,
        ip_prefixes: list[Node] | None = None,
        prefix_member_type: PrefixMemberType | None = None,
        branch_names: list[str] | None = None,
    ) -> int:
        children = await self.get_children(
            ip_prefixes=ip_prefixes, prefix_member_type=prefix_member_type, branch_names=branch_names
        )
        return len(children)

    async def _get_prefix_use_fraction(
        self, ip_prefixes: list[Node] | None = None, branch_names: list[str] | None = None
    ) -> tuple[int, int]:
        total_prefix_space = 0
        total_used_space = 0
        if ip_prefixes is None:
            ip_prefixes = self.ip_prefixes
        for ip_prefix in ip_prefixes:
            total_prefix_space += get_prefix_space(ip_prefix=ip_prefix)
            max_prefixlen = ip_prefix.prefix.obj.max_prefixlen  # type: ignore[attr-defined]
            children = await self.get_children(
                ip_prefixes=[ip_prefix], prefix_member_type=PrefixMemberType.PREFIX, branch_names=branch_names
            )
            for child in children:
                total_used_space += 2 ** (max_prefixlen - child.prefixlen)
        return total_used_space, total_prefix_space

    async def _get_address_use_fraction(
        self, ip_prefixes: list[Node] | None = None, branch_names: list[str] | None = None
    ) -> tuple[int, int]:
        total_prefix_space = 0
        if ip_prefixes is None:
            ip_prefixes = self.ip_prefixes
        for ip_prefix in ip_prefixes:
            total_prefix_space += get_prefix_space(ip_prefix=ip_prefix)
        total_used_space = await self.get_num_children_in_use(
            ip_prefixes=ip_prefixes, prefix_member_type=PrefixMemberType.ADDRESS, branch_names=branch_names
        )
        return total_used_space, total_prefix_space

    async def get_use_percentage(
        self, ip_prefixes: list[Node] | None = None, branch_names: list[str] | None = None
    ) -> float:
        grand_total_used, grand_total_space = 0, 0
        address_prefixes, prefix_prefixes = [], []
        if ip_prefixes is None:
            ip_prefixes = self.ip_prefixes
        for ip_prefix in ip_prefixes:
            if ip_prefix.member_type.value == PrefixMemberType.ADDRESS.value:  # type: ignore[union-attr,attr-defined]
                address_prefixes.append(ip_prefix)
            else:
                prefix_prefixes.append(ip_prefix)
        if address_prefixes:
            address_total_used, address_total_space = await self._get_address_use_fraction(
                ip_prefixes=address_prefixes, branch_names=branch_names
            )
            grand_total_used += address_total_used
            grand_total_space += address_total_space
        if prefix_prefixes:
            prefix_total_used, prefix_total_space = await self._get_prefix_use_fraction(
                ip_prefixes=prefix_prefixes, branch_names=branch_names
            )
            grand_total_used += prefix_total_used
            grand_total_space += prefix_total_space
        if grand_total_space == 0:
            return 0.0
        return min((grand_total_used / grand_total_space) * 100, 100)



from __future__ import annotations

from typing import Optional

import typer
from rich.console import Console
from rich.table import Table

from ..async_typer import AsyncTyper
from ..task.manager import TaskFilter
from ..task.models import Task, TaskState
from .client import initialize_client
from .parameters import CONFIG_PARAM
from .utils import catch_exception, init_logging

app = AsyncTyper()
console = Console()


@app.callback()
def callback() -> None:
    """Manage Infrahub tasks."""


def _parse_states(states: list[str] | None) -> list[TaskState] | None:
    if not states:
        return None

    parsed_states: list[TaskState] = []
    for state in states:
        normalized_state = state.strip().upper()
        try:
            parsed_states.append(TaskState(normalized_state))
        except ValueError as exc:  # pragma: no cover - typer will surface this as CLI error
            raise typer.BadParameter(
                f"Unsupported state '{state}'. Available states: {', '.join(item.value.lower() for item in TaskState)}"
            ) from exc

    return parsed_states


def _render_table(tasks: list[Task]) -> None:
    table = Table(title="Infrahub Tasks", box=None)
    table.add_column("ID", style="cyan", overflow="fold")
    table.add_column("Title", style="magenta", overflow="fold")
    table.add_column("State", style="green")
    table.add_column("Progress", justify="right")
    table.add_column("Workflow", overflow="fold")
    table.add_column("Branch", overflow="fold")
    table.add_column("Updated")

    if not tasks:
        table.add_row("-", "No tasks found", "-", "-", "-", "-", "-")
        console.print(table)
        return

    for task in tasks:
        progress = f"{task.progress:.0%}" if task.progress is not None else "-"
        table.add_row(
            task.id,
            task.title,
            task.state.value,
            progress,
            task.workflow or "-",
            task.branch or "-",
            task.updated_at.isoformat(),
        )

    console.print(table)


@app.command(name="list")
@catch_exception(console=console)
async def list_tasks(
    state: list[str] = typer.Option(
        None, "--state", "-s", help="Filter by task state. Can be provided multiple times."
    ),
    limit: Optional[int] = typer.Option(None, help="Maximum number of tasks to retrieve."),
    offset: Optional[int] = typer.Option(None, help="Offset for pagination."),
    include_related_nodes: bool = typer.Option(False, help="Include related nodes in the output."),
    include_logs: bool = typer.Option(False, help="Include task logs in the output."),
    json_output: bool = typer.Option(False, "--json", help="Output the result as JSON."),
    debug: bool = False,
    _: str = CONFIG_PARAM,
) -> None:
    """List Infrahub tasks."""

    init_logging(debug=debug)

    client = initialize_client()
    filters = TaskFilter()
    parsed_states = _parse_states(state)
    if parsed_states:
        filters.state = parsed_states

    tasks = await client.task.filter(
        filter=filters,
        limit=limit,
        offset=offset,
        include_related_nodes=include_related_nodes,
        include_logs=include_logs,
    )

    if json_output:
        console.print_json(
            data=[task.model_dump(mode="json") for task in tasks], indent=2, sort_keys=True, highlight=False
        )
        return

    _render_table(tasks)



from __future__ import annotations

from asyncio import run as aiorun
from pathlib import Path
from typing import Optional

import typer
from rich.console import Console

from ..ctl.client import initialize_client
from ..transfer.exceptions import TransferError
from ..transfer.importer.json import LineDelimitedJSONImporter
from ..transfer.schema_sorter import InfrahubSchemaTopologicalSorter
from .parameters import CONFIG_PARAM


def local_directory() -> Path:
    # We use a function here to avoid failure when generating the documentation due to directory name
    return Path().resolve()


def load(
    directory: Path = typer.Option(local_directory, help="Directory path of exported data"),
    continue_on_error: bool = typer.Option(
        False, help="Allow exceptions during loading and display them when complete"
    ),
    quiet: bool = typer.Option(False, help="No console output"),
    _: str = CONFIG_PARAM,
    branch: str = typer.Option(None, help="Branch from which to export"),
    concurrent: Optional[int] = typer.Option(
        None,
        help="Maximum number of requests to execute at the same time.",
        envvar="INFRAHUB_MAX_CONCURRENT_EXECUTION",
    ),
    timeout: int = typer.Option(60, help="Timeout in sec", envvar="INFRAHUB_TIMEOUT"),
) -> None:
    """Import nodes and their relationships into the database."""
    console = Console()

    client = initialize_client(
        branch=branch, timeout=timeout, max_concurrent_execution=concurrent, retry_on_failure=True
    )

    importer = LineDelimitedJSONImporter(
        client,
        InfrahubSchemaTopologicalSorter(),
        continue_on_error=continue_on_error,
        console=Console() if not quiet else None,
    )
    try:
        aiorun(importer.import_data(import_directory=directory, branch=branch))
    except TransferError as exc:
        console.print(f"[red]{exc}")
        raise typer.Exit(1)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Self

from infrahub_sdk.graphql import Query
from pydantic import BaseModel, Field

from infrahub.core.constants import RelationshipCardinality
from infrahub.core.registry import registry
from infrahub.core.schema import NodeSchema  # noqa: TC001
from infrahub.events import NodeUpdatedEvent
from infrahub.trigger.constants import NAME_SEPARATOR
from infrahub.trigger.models import (
    EventTrigger,
    ExecuteWorkflow,
    TriggerBranchDefinition,
    TriggerType,
)
from infrahub.workflows.catalogue import HFID_PROCESS

if TYPE_CHECKING:
    from infrahub.core.schema.schema_branch_hfid import HFIDs, RelationshipTriggers


@dataclass
class AttributeTarget:
    hash: str
    fields: set[str]


class HFIDTriggerDefinition(TriggerBranchDefinition):
    type: TriggerType = TriggerType.HUMAN_FRIENDLY_ID
    hfid_hash: str
    target_kind: str | None = Field(default=None)

    def get_description(self) -> str:
        return f"{super().get_description()} | hash:{self.hfid_hash}"

    @classmethod
    def from_schema_hfids(
        cls,
        branch: str,
        hfids: HFIDs,
        branches_out_of_scope: list[str] | None = None,
    ) -> list[HFIDTriggerDefinition]:
        """
        This function is used to create a trigger definition for a display labels of type Jinja2.
        """

        definitions: list[HFIDTriggerDefinition] = []

        for node_kind, hfid_definition in hfids.get_template_nodes().items():
            definitions.append(
                cls.new(
                    branch=branch,
                    node_kind=node_kind,
                    target_kind=node_kind,
                    fields=[
                        "_trigger_placeholder"
                    ],  # Triggers for the nodes themselves are only used to determine if all nodes should be regenerated
                    hfid_hash=hfid_definition.get_hash(),
                    branches_out_of_scope=branches_out_of_scope,
                )
            )

        for related_kind, relationship_trigger in hfids.get_related_trigger_nodes().items():
            definitions.extend(
                cls.from_related_node(
                    branch=branch,
                    related_kind=related_kind,
                    relationship_trigger=relationship_trigger,
                    hfids=hfids,
                    branches_out_of_scope=branches_out_of_scope,
                )
            )

        return definitions

    @classmethod
    def from_related_node(
        cls,
        branch: str,
        related_kind: str,
        relationship_trigger: RelationshipTriggers,
        hfids: HFIDs,
        branches_out_of_scope: list[str] | None = None,
    ) -> list[HFIDTriggerDefinition]:
        targets_by_attribute: dict[str, AttributeTarget] = {}
        definitions: list[HFIDTriggerDefinition] = []
        for attribute, relationship_identifiers in relationship_trigger.attributes.items():
            for relationship_identifier in relationship_identifiers:
                actual_node = hfids.get_node_definition(kind=relationship_identifier.kind)
                if relationship_identifier.kind not in targets_by_attribute:
                    targets_by_attribute[relationship_identifier.kind] = AttributeTarget(
                        actual_node.get_hash(), fields=set()
                    )
                targets_by_attribute[relationship_identifier.kind].fields.add(attribute)

        for target_kind, attribute_target in targets_by_attribute.items():
            definitions.append(
                cls.new(
                    branch=branch,
                    node_kind=related_kind,
                    target_kind=target_kind,
                    fields=sorted(attribute_target.fields),
                    hfid_hash=attribute_target.hash,
                    branches_out_of_scope=branches_out_of_scope,
                )
            )

        return definitions

    @classmethod
    def new(
        cls,
        branch: str,
        node_kind: str,
        target_kind: str,
        hfid_hash: str,
        fields: list[str],
        branches_out_of_scope: list[str] | None = None,
    ) -> Self:
        event_trigger = EventTrigger()
        event_trigger.events.add(NodeUpdatedEvent.event_name)
        event_trigger.match = {"infrahub.node.kind": node_kind}
        if branches_out_of_scope:
            event_trigger.match["infrahub.branch.name"] = [f"!{branch}" for branch in branches_out_of_scope]
        elif not branches_out_of_scope and branch != registry.default_branch:
            event_trigger.match["infrahub.branch.name"] = branch

        event_trigger.match_related = {
            "prefect.resource.role": ["infrahub.node.attribute_update", "infrahub.node.relationship_update"],
            "infrahub.field.name": fields,
        }

        workflow = ExecuteWorkflow(
            workflow=HFID_PROCESS,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "node_kind": node_kind,
                "object_id": "{{ event.resource['infrahub.node.id'] }}",
                "target_kind": target_kind,
                "context": {
                    "__prefect_kind": "json",
                    "value": {
                        "__prefect_kind": "jinja",
                        "template": "{{ event.payload['context'] | tojson }}",
                    },
                },
            },
        )

        trigger_definition_target_kind = target_kind if target_kind == node_kind else None

        return cls(
            name=f"{target_kind}{NAME_SEPARATOR}by{NAME_SEPARATOR}{node_kind}",
            hfid_hash=hfid_hash,
            branch=branch,
            trigger=event_trigger,
            actions=[workflow],
            target_kind=trigger_definition_target_kind,
        )


class HFIDGraphQLResponse(BaseModel):
    node_id: str
    hfid_value: list[str] | None = Field(default=None)
    variables: dict[str, str] = Field(default_factory=dict)


class HFIDGraphQL(BaseModel):
    filter_key: str
    node_schema: NodeSchema = Field(..., description="The node kind where the computed attribute is defined")
    variables: list[str] = Field(..., description="The list of variable names used within the computed attribute")

    def render_graphql_query(self, filter_id: str) -> str:
        query_fields = self.query_fields
        query_fields["id"] = None
        query_fields["hfid"] = None
        query = Query(
            name="HFIDFilter",
            query={
                self.node_schema.kind: {
                    "@filters": {self.filter_key: filter_id},
                    "edges": {"node": query_fields},
                }
            },
        )

        return query.render()

    @property
    def query_fields(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        for variable in self.variables:
            field_name, remainder = variable.split("__", maxsplit=1)
            if field_name in self.node_schema.attribute_names:
                output[field_name] = {remainder: None}
            elif field_name in self.node_schema.relationship_names:
                related_attribute, related_value = remainder.split("__", maxsplit=1)
                relationship = self.node_schema.get_relationship(name=field_name)
                if relationship.cardinality == RelationshipCardinality.ONE:
                    if field_name not in output:
                        output[field_name] = {"node": {}}
                    output[field_name]["node"][related_attribute] = {related_value: None}
        return output

    def parse_response(self, response: dict[str, Any]) -> list[HFIDGraphQLResponse]:
        rendered_response: list[HFIDGraphQLResponse] = []
        if kind_payload := response.get(self.node_schema.kind):
            edges = kind_payload.get("edges", [])
            for node in edges:
                if node_response := self.to_node_response(node_dict=node):
                    rendered_response.append(node_response)
        return rendered_response

    def to_node_response(self, node_dict: dict[str, Any]) -> HFIDGraphQLResponse | None:
        if node := node_dict.get("node"):
            node_id = node.get("id")
        else:
            return None

        hfid = node.get("hfid")
        response = HFIDGraphQLResponse(node_id=node_id, hfid_value=hfid)
        for variable in self.variables:
            field_name, remainder = variable.split("__", maxsplit=1)
            # response.variables[variable] = None
            if field_content := node.get(field_name):
                if field_name in self.node_schema.attribute_names:
                    response.variables[variable] = str(field_content.get(remainder, ""))
                elif field_name in self.node_schema.relationship_names:
                    relationship = self.node_schema.get_relationship(name=field_name)
                    if relationship.cardinality == RelationshipCardinality.ONE:
                        related_attribute, related_value = remainder.split("__", maxsplit=1)
                        node_content = field_content.get("node") or {}
                        related_attribute_content = node_content.get(related_attribute) or {}
                        response.variables[variable] = str(related_attribute_content.get(related_value, ""))

        return response



from __future__ import annotations

from enum import Flag, auto

from infrahub.core.constants import infrahubkind as InfrahubKind  # noqa: N812
from infrahub.exceptions import ValidationError
from infrahub.utils import InfrahubNumberEnum, InfrahubStringEnum

from .schema import FlagProperty, NodeProperty, SchemaElementPathType, UpdateSupport, UpdateValidationErrorType

__all__ = [
    "FlagProperty",
    "InfrahubKind",
    "NodeProperty",
    "SchemaElementPathType",
    "UpdateSupport",
    "UpdateValidationErrorType",
    "ValidationError",
]


GLOBAL_BRANCH_NAME = "-global-"

DEFAULT_IP_NAMESPACE = "default"

RESERVED_BRANCH_NAMES = [GLOBAL_BRANCH_NAME]

RESERVED_ATTR_REL_NAMES = [
    "any",
    "attribute",
    "attributes",
    "attr",
    "attrs",
    "relationship",
    "relationships",
    "rel",
    "rels",
    "save",
    "hfid",
    "process_pools",
]

RESERVED_ATTR_GEN_NAMES = ["type"]

NULL_VALUE = "NULL"

EVENT_NAMESPACE = "infrahub"


class EventType(InfrahubStringEnum):
    BRANCH_CREATED = f"{EVENT_NAMESPACE}.branch.created"
    BRANCH_DELETED = f"{EVENT_NAMESPACE}.branch.deleted"
    BRANCH_MERGED = f"{EVENT_NAMESPACE}.branch.merged"
    BRANCH_MIGRATED = f"{EVENT_NAMESPACE}.branch.migrated"
    BRANCH_REBASED = f"{EVENT_NAMESPACE}.branch.rebased"

    SCHEMA_UPDATED = f"{EVENT_NAMESPACE}.schema.updated"

    NODE_CREATED = f"{EVENT_NAMESPACE}.node.created"
    NODE_UPDATED = f"{EVENT_NAMESPACE}.node.updated"
    NODE_DELETED = f"{EVENT_NAMESPACE}.node.deleted"

    GROUP_MEMBER_ADDED = f"{EVENT_NAMESPACE}.group.member_added"
    GROUP_MEMBER_REMOVED = f"{EVENT_NAMESPACE}.group.member_removed"

    PROPOSED_CHANGE_MERGED = f"{EVENT_NAMESPACE}.proposed_change.merged"
    PROPOSED_CHANGE_REVIEW_REQUESTED = f"{EVENT_NAMESPACE}.proposed_change.review_requested"
    PROPOSED_CHANGE_APPROVED = f"{EVENT_NAMESPACE}.proposed_change.approved"
    PROPOSED_CHANGE_REJECTED = f"{EVENT_NAMESPACE}.proposed_change.rejected"
    PROPOSED_CHANGE_APPROVAL_REVOKED = f"{EVENT_NAMESPACE}.proposed_change.approval_revoked"
    PROPOSED_CHANGE_APPROVALS_REVOKED = f"{EVENT_NAMESPACE}.proposed_change.approvals_revoked"
    PROPOSED_CHANGE_REJECTION_REVOKED = f"{EVENT_NAMESPACE}.proposed_change.rejection_revoked"
    PROPOSED_CHANGE_THREAD_CREATED = f"{EVENT_NAMESPACE}.proposed_change_thread.created"
    PROPOSED_CHANGE_THREAD_UPDATED = f"{EVENT_NAMESPACE}.proposed_change_thread.updated"

    REPOSITORY_UPDATE_COMMIT = f"{EVENT_NAMESPACE}.repository.update_commit"

    ARTIFACT_CREATED = f"{EVENT_NAMESPACE}.artifact.created"
    ARTIFACT_UPDATED = f"{EVENT_NAMESPACE}.artifact.updated"

    VALIDATOR_STARTED = f"{EVENT_NAMESPACE}.validator.started"
    VALIDATOR_PASSED = f"{EVENT_NAMESPACE}.validator.passed"
    VALIDATOR_FAILED = f"{EVENT_NAMESPACE}.validator.failed"


class PermissionLevel(Flag):
    READ = 1
    WRITE = 2
    ADMIN = 3
    DEFAULT = 0


class GlobalPermissions(InfrahubStringEnum):
    EDIT_DEFAULT_BRANCH = "edit_default_branch"
    SUPER_ADMIN = "super_admin"
    MERGE_BRANCH = "merge_branch"
    MERGE_PROPOSED_CHANGE = "merge_proposed_change"
    REVIEW_PROPOSED_CHANGE = "review_proposed_change"
    MANAGE_SCHEMA = "manage_schema"
    MANAGE_ACCOUNTS = "manage_accounts"
    MANAGE_PERMISSIONS = "manage_permissions"
    MANAGE_REPOSITORIES = "manage_repositories"
    OVERRIDE_CONTEXT = "override_context"
    UPDATE_OBJECT_HFID_DISPLAY_LABEL = "update_object_hfid_display_label"


class PermissionAction(InfrahubStringEnum):
    ANY = "any"
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"
    VIEW = "view"


class PermissionDecision(InfrahubNumberEnum):
    DENY = 1
    ALLOW_DEFAULT = 2
    ALLOW_OTHER = 4
    ALLOW_ALL = 6


class AccountType(InfrahubStringEnum):
    USER = "User"
    SCRIPT = "Script"
    BOT = "Bot"
    Git = "Git"


class NumberPoolType(InfrahubStringEnum):
    USER = "User"
    SCHEMA = "Schema"


class AccountStatus(InfrahubStringEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"


class ArtifactStatus(InfrahubStringEnum):
    ERROR = "Error"
    PENDING = "Pending"
    PROCESSING = "Processing"
    READY = "Ready"


class BranchSupportType(InfrahubStringEnum):
    AWARE = "aware"
    AGNOSTIC = "agnostic"
    LOCAL = "local"


class ComputedAttributeKind(InfrahubStringEnum):
    USER = "User"
    JINJA2 = "Jinja2"
    TRANSFORM_PYTHON = "TransformPython"


class BranchConflictKeep(InfrahubStringEnum):
    TARGET = "target"
    SOURCE = "source"


class AllowOverrideType(InfrahubStringEnum):
    NONE = "none"
    ANY = "any"


class RepositoryObjects(InfrahubStringEnum):
    OBJECT = "object"
    MENU = "menu"


class ContentType(InfrahubStringEnum):
    APPLICATION_JSON = "application/json"
    APPLICATION_YAML = "application/yaml"
    APPLICATION_XML = "application/xml"
    APPLICATION_HCL = "application/hcl"
    TEXT_PLAIN = "text/plain"
    TEXT_MARKDOWN = "text/markdown"
    TEXT_CSV = "text/csv"
    IMAGE_SVG = "image/svg+xml"


class CheckType(InfrahubStringEnum):
    ARTIFACT = "artifact"
    DATA = "data"
    GENERATOR = "generator"
    REPOSITORY = "repository"
    SCHEMA = "schema"
    TEST = "test"
    USER = "user"
    ALL = "all"


class RepositoryInternalStatus(InfrahubStringEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    STAGING = "staging"


class RepositorySyncStatus(InfrahubStringEnum):
    UNKNOWN = "unknown"
    IN_SYNC = "in-sync"
    ERROR_IMPORT = "error-import"
    SYNCING = "syncing"


class RepositoryOperationalStatus(InfrahubStringEnum):
    UNKNOWN = "unknown"
    ERROR_CRED = "error-cred"
    ERROR_CONNECTION = "error-connection"
    ERROR = "error"
    ONLINE = "online"


class DiffAction(InfrahubStringEnum):
    ADDED = "added"
    REMOVED = "removed"
    UPDATED = "updated"
    UNCHANGED = "unchanged"


class GeneratorInstanceStatus(InfrahubStringEnum):
    ERROR = "Error"
    PENDING = "Pending"
    PROCESSING = "Processing"
    READY = "Ready"


class MutationAction(InfrahubStringEnum):
    CREATED = "created"
    DELETED = "deleted"
    UPDATED = "updated"
    UNDEFINED = "undefined"

    @classmethod
    def from_diff_action(cls, diff_action: DiffAction) -> MutationAction:
        match diff_action:
            case DiffAction.ADDED:
                return MutationAction.CREATED
            case DiffAction.REMOVED:
                return MutationAction.DELETED
            case DiffAction.UPDATED:
                return MutationAction.UPDATED
            case DiffAction.UNCHANGED:
                return MutationAction.UNDEFINED


class PathResourceType(InfrahubStringEnum):
    SCHEMA = "schema"
    DATA = "data"
    FILE = "file"


class SchemaPathType(InfrahubStringEnum):
    NODE = "node"
    ATTRIBUTE = "attribute"
    RELATIONSHIP = "relationship"


class PathType(InfrahubStringEnum):
    NODE = "node"
    ATTRIBUTE = "attribute"
    RELATIONSHIP_ONE = "relationship_one"
    RELATIONSHIP_MANY = "relationship_many"

    @classmethod
    def from_relationship(cls, relationship: RelationshipCardinality) -> PathType:
        if relationship == RelationshipCardinality.ONE:
            return cls("relationship_one")

        return cls("relationship_many")


class HashableModelState(InfrahubStringEnum):
    PRESENT = "present"
    ABSENT = "absent"


class RelationshipCardinality(InfrahubStringEnum):
    ONE = "one"
    MANY = "many"


class RelationshipKind(InfrahubStringEnum):
    GENERIC = "Generic"
    ATTRIBUTE = "Attribute"
    COMPONENT = "Component"
    PARENT = "Parent"
    GROUP = "Group"
    HIERARCHY = "Hierarchy"
    PROFILE = "Profile"
    TEMPLATE = "Template"


class RelationshipStatus(InfrahubStringEnum):
    ACTIVE = "active"
    DELETED = "deleted"


class RelationshipDirection(InfrahubStringEnum):
    BIDIR = "bidirectional"
    OUTBOUND = "outbound"
    INBOUND = "inbound"

    @property
    def neighbor_direction(self) -> RelationshipDirection:
        NEIGHBOR_MAP = {
            RelationshipDirection.BIDIR: RelationshipDirection.BIDIR,
            RelationshipDirection.INBOUND: RelationshipDirection.OUTBOUND,
            RelationshipDirection.OUTBOUND: RelationshipDirection.INBOUND,
        }
        return NEIGHBOR_MAP[self]


class RelationshipHierarchyDirection(InfrahubStringEnum):
    ANCESTORS = "ancestors"
    DESCENDANTS = "descendants"


class RelationshipDeleteBehavior(InfrahubStringEnum):
    NO_ACTION = "no-action"
    CASCADE = "cascade"


class Severity(InfrahubStringEnum):
    SUCCESS = "success"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


class TaskConclusion(InfrahubStringEnum):
    UNKNOWN = "unknown"
    FAILURE = "failure"
    SUCCESS = "success"


class ValidatorConclusion(InfrahubStringEnum):
    UNKNOWN = "unknown"
    FAILURE = "failure"
    SUCCESS = "success"


class ValidatorState(InfrahubStringEnum):
    QUEUED = "queued"
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"


class AttributeDBNodeType(Flag):
    DEFAULT = auto()
    INDEX_ONLY = auto()
    IPHOST_ONLY = auto()
    IPNETWORK_ONLY = auto()
    INDEXED = DEFAULT | INDEX_ONLY
    IPHOST = DEFAULT | INDEX_ONLY | IPHOST_ONLY
    IPNETWORK = DEFAULT | INDEX_ONLY | IPNETWORK_ONLY


RESTRICTED_NAMESPACES: list[str] = [
    "Account",
    "Branch",
    # "Builtin",
    "Core",
    "Deprecated",
    "Diff",
    "Infrahub",
    "Internal",
    "Lineage",
    "Schema",
    "Profile",
    "Template",
]

NODE_NAME_REGEX = r"^[A-Z][a-zA-Z0-9]+$"
DEFAULT_NAME_MIN_LENGTH = 2
NAME_REGEX = r"^[a-z0-9\_]+$"
NAME_REGEX_OR_EMPTY = r"^[a-z0-9\_]*$"
DEFAULT_DESCRIPTION_LENGTH = 128

DEFAULT_NAME_MAX_LENGTH = 32
DEFAULT_LABEL_MAX_LENGTH = 64
DEFAULT_KIND_MIN_LENGTH = 3
DEFAULT_KIND_MAX_LENGTH = 32
NAMESPACE_REGEX = r"^[A-Z][a-z0-9]+$"
NODE_KIND_REGEX = r"^[A-Z][a-zA-Z0-9]+$"
DEFAULT_REL_IDENTIFIER_LENGTH = 128

OBJECT_TEMPLATE_RELATIONSHIP_NAME = "object_template"
OBJECT_TEMPLATE_NAME_ATTR = "template_name"
PROFILE_NODE_RELATIONSHIP_IDENTIFIER = "node__profile"



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from infrahub.core.constants import RepositoryInternalStatus
from infrahub.core.constants.infrahubkind import READONLYREPOSITORY, REPOSITORY
from infrahub.core.protocols import CoreGenericRepository, CoreReadOnlyRepository, CoreRepository
from infrahub.exceptions import ValidationError
from infrahub.git.models import GitRepositoryAdd, GitRepositoryAddReadOnly
from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.message_bus.messages.git_repository_connectivity import GitRepositoryConnectivityResponse
from infrahub.workflows.catalogue import GIT_REPOSITORY_ADD, GIT_REPOSITORY_ADD_READ_ONLY

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.context import InfrahubContext
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase
    from infrahub.services import InfrahubServices

log = get_logger()


class RepositoryFinalizer:
    def __init__(
        self,
        account_session: AccountSession,
        services: InfrahubServices,
        context: InfrahubContext,
    ) -> None:
        self.account_session = account_session
        self.services = services
        self.context = context

    async def post_create(
        self,
        obj: CoreGenericRepository,
        branch: Branch,
        db: InfrahubDatabase,
        delete_on_connectivity_failure: bool = True,
    ) -> None:
        """
        Method meant to be called after a repository has been created in the database.
        It mainly checks the connectivity to the remote repository and submit the workflow to create the repository in the local filesystem.
        """

        # If the connectivity is not good, we remove the repository to allow the user to add a new one
        if delete_on_connectivity_failure:
            message = messages.GitRepositoryConnectivity(
                repository_name=obj.name.value,
                repository_location=obj.location.value,
            )
            response = await self.services.message_bus.rpc(
                message=message, response_class=GitRepositoryConnectivityResponse
            )

            if response.data.success is False:
                await obj.delete(db=db)
                raise ValidationError(response.data.message)

        # If we are in the default branch, we set the sync status to Active
        # If we are in another branch, we set the sync status to Staging
        if branch.is_default:
            obj.internal_status.value = RepositoryInternalStatus.ACTIVE.value
        else:
            obj.internal_status.value = RepositoryInternalStatus.STAGING.value
        await obj.save(db=db)

        # Create the new repository in the filesystem.
        log.info("create_repository", name=obj.name.value)
        authenticated_user = None
        if self.account_session and self.account_session.authenticated:
            authenticated_user = self.account_session.account_id

        if obj.get_kind() == READONLYREPOSITORY:
            obj = cast(CoreReadOnlyRepository, obj)
            model = GitRepositoryAddReadOnly(
                repository_id=obj.id,
                repository_name=obj.name.value,
                location=obj.location.value,
                ref=obj.ref.value,
                infrahub_branch_name=branch.name,
                infrahub_branch_id=str(branch.get_uuid()),
                internal_status=obj.internal_status.value,
                created_by=authenticated_user,
            )
            await self.services.workflow.submit_workflow(
                workflow=GIT_REPOSITORY_ADD_READ_ONLY,
                context=self.context,
                parameters={"model": model},
            )

        elif obj.get_kind() == REPOSITORY:
            obj = cast(CoreRepository, obj)
            git_repo_add_model = GitRepositoryAdd(
                repository_id=obj.id,
                repository_name=obj.name.value,
                location=obj.location.value,
                default_branch_name=obj.default_branch.value,
                infrahub_branch_name=branch.name,
                infrahub_branch_id=str(branch.get_uuid()),
                internal_status=obj.internal_status.value,
                created_by=authenticated_user,
            )

            await self.services.workflow.submit_workflow(
                workflow=GIT_REPOSITORY_ADD,
                context=self.context,
                parameters={"model": git_repo_add_model},
            )
        else:
            raise ValueError(f"Unknown repository kind: {obj.get_kind()}")



from __future__ import annotations

from typing import cast

from infrahub_sdk.exceptions import URLNotFoundError
from prefect import flow
from prefect.logging import get_run_logger

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core.registry import registry
from infrahub.events import BranchDeletedEvent
from infrahub.trigger.models import TriggerSetupReport, TriggerType
from infrahub.trigger.setup import setup_triggers_specific
from infrahub.workers.dependencies import get_client, get_component, get_database, get_workflow
from infrahub.workflows.catalogue import HFID_PROCESS, TRIGGER_UPDATE_HFID
from infrahub.workflows.utils import add_tags, wait_for_schema_to_converge

from .gather import gather_trigger_hfid
from .models import HFIDGraphQL, HFIDGraphQLResponse, HFIDTriggerDefinition

UPDATE_HFID = """
mutation UpdateHFID(
    $id: String!,
    $kind: String!,
    $value: [String!]!
  ) {
  InfrahubUpdateHFID(
    data: {id: $id, value: $value, kind: $kind}
  ) {
    ok
  }
}
"""


@flow(
    name="hfid-update-value",
    flow_run_name="Update value for hfid on {node_kind}",
)
async def hfid_update_value(
    branch_name: str,
    obj: HFIDGraphQLResponse,
    node_kind: str,
    hfid_definition: list[str],
) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name], nodes=[obj.node_id], db_change=True)

    rendered_hfid: list[str] = []
    for hfid_component in hfid_definition:
        if hfid_component in obj.variables:
            rendered_hfid.append(obj.variables[hfid_component])
    # value = await template.render(variables=obj.variables)
    if rendered_hfid == obj.hfid_value:
        log.debug(f"Ignoring to update {obj} with existing value on human_friendly_id={obj.hfid_value}")
        return

    try:
        await client.execute_graphql(
            query=UPDATE_HFID,
            variables={"id": obj.node_id, "kind": node_kind, "value": rendered_hfid},
            branch_name=branch_name,
        )
        log.info(f"Updating {node_kind}.human_friendly_id='{rendered_hfid}' ({obj.node_id})")
    except URLNotFoundError:
        log.warning(
            f"Updating {node_kind}.human_friendly_id='{rendered_hfid}' ({obj.node_id}) failed for branch {branch_name} (branch not found)"
        )


@flow(
    name="hfid-process",
    flow_run_name="Process human friendly ids for {target_kind}",
)
async def process_hfid(
    branch_name: str,
    node_kind: str,
    object_id: str,
    target_kind: str,
    context: InfrahubContext,  # noqa: ARG001
) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name])

    target_schema = branch_name if branch_name in registry.get_altered_schema_branches() else registry.default_branch
    schema_branch = registry.schema.get_schema_branch(name=target_schema)
    node_schema = schema_branch.get_node(name=target_kind, duplicate=False)

    if node_kind == target_kind:
        hfid_definition = schema_branch.hfids.get_node_definition(kind=node_kind)
    else:
        hfid_definition = schema_branch.hfids.get_related_definition(related_kind=node_kind, target_kind=target_kind)

    # jinja_template = Jinja2Template(template=display_label_template.template)
    # variables = jinja_template.get_variables()
    hfid_graphql = HFIDGraphQL(
        node_schema=node_schema, variables=hfid_definition.hfid, filter_key=hfid_definition.filter_key
    )

    query = hfid_graphql.render_graphql_query(filter_id=object_id)
    response = await client.execute_graphql(query=query, branch_name=branch_name)
    update_candidates = hfid_graphql.parse_response(response=response)

    if not update_candidates:
        log.debug("No nodes found that requires updates")
        return

    batch = await client.create_batch()
    for node in update_candidates:
        batch.add(
            task=hfid_update_value,
            branch_name=branch_name,
            obj=node,
            node_kind=node_schema.kind,
            hfid_definition=hfid_definition.hfid,
        )

    _ = [response async for _, response in batch.execute()]


@flow(name="hfid-setup", flow_run_name="Setup human friendly ids in task-manager")
async def hfid_setup(context: InfrahubContext, branch_name: str | None = None, event_name: str | None = None) -> None:
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        if branch_name:
            await add_tags(branches=[branch_name])
            component = await get_component()
            await wait_for_schema_to_converge(branch_name=branch_name, component=component, db=db, log=log)

        report: TriggerSetupReport = await setup_triggers_specific(
            gatherer=gather_trigger_hfid, trigger_type=TriggerType.HUMAN_FRIENDLY_ID
        )  # type: ignore[misc]

        # Configure all DisplayLabelTriggerDefinitions in Prefect
        hfid_reports = [cast(HFIDTriggerDefinition, entry) for entry in report.updated + report.created]
        direct_target_triggers = [hfid_report for hfid_report in hfid_reports if hfid_report.target_kind]

        for display_report in direct_target_triggers:
            if event_name != BranchDeletedEvent.event_name and display_report.branch == branch_name:
                await get_workflow().submit_workflow(
                    workflow=TRIGGER_UPDATE_HFID,
                    context=context,
                    parameters={
                        "branch_name": display_report.branch,
                        "kind": display_report.target_kind,
                    },
                )

        log.info(f"{report.in_use_count} HFID automation configurations completed")


@flow(
    name="trigger-update-hfid",
    flow_run_name="Trigger updates for HFID for {kind}",
)
async def trigger_update_hfid(
    branch_name: str,
    kind: str,
    context: InfrahubContext,
) -> None:
    await add_tags(branches=[branch_name])

    client = get_client()

    # NOTE we only need the id of the nodes, this query will still query for the HFID
    node_schema = registry.schema.get_node_schema(name=kind, branch=branch_name)
    nodes = await client.all(
        kind=kind,
        branch=branch_name,
        exclude=node_schema.attribute_names + node_schema.relationship_names,
        populate_store=False,
    )

    for node in nodes:
        await get_workflow().submit_workflow(
            workflow=HFID_PROCESS,
            context=context,
            parameters={
                "branch_name": branch_name,
                "node_kind": kind,
                "target_kind": kind,
                "object_id": node.id,
                "context": context,
            },
        )



from fastapi import APIRouter
from fastapi.param_functions import Depends
from prefect.server.database import PrefectDBInterface, provide_database_interface

from .database import query_events
from .models import InfrahubEventfilterInput, InfrahubEventPage

router = APIRouter(prefix="/events", tags=["Infrahub"])


@router.post(
    "/filter",
)
async def read_events(
    event_filter: InfrahubEventfilterInput,
    db: PrefectDBInterface = Depends(provide_database_interface),  # noqa: B008
) -> InfrahubEventPage:
    event_filter.filter.set_prefix()

    async with db.session_context() as session:
        events, total = await query_events(
            session=session, filter=event_filter.filter, page_size=event_filter.limit, offset=event_filter.offset
        )

        return InfrahubEventPage(
            events=events,
            total=total,
        )



from infrahub.core.diff.diff_locker import DiffLocker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffLockerDependency(DependencyBuilder[DiffLocker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffLocker:  # noqa: ARG003
        return DiffLocker()



from infrahub.core.diff.conflict_transferer import DiffConflictTransferer
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .combiner import DiffCombinerDependency


class DiffConflictTransfererDependency(DependencyBuilder[DiffConflictTransferer]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffConflictTransferer:
        return DiffConflictTransferer(diff_combiner=DiffCombinerDependency.build(context=context))



from dataclasses import dataclass
from typing import Any

from infrahub.core.constants import NULL_VALUE
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


@dataclass
class ProfileData:
    uuid: str
    priority: float | int
    attribute_values: dict[str, Any]


class GetProfileDataQuery(Query):
    type: QueryType = QueryType.READ
    insert_return: bool = False

    def __init__(self, *args: Any, profile_ids: list[str], attr_names: list[str], **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.profile_ids = profile_ids
        self.attr_names = attr_names

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)
        self.params["profile_ids"] = self.profile_ids
        self.params["attr_names"] = self.attr_names + ["profile_priority"]

        query = """
// --------------
// get the Profile nodes
// --------------
MATCH (profile:Node)
WHERE profile.uuid IN $profile_ids
// --------------
// make sure we only use the active ones
// --------------
CALL (profile) {
    MATCH (profile)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    RETURN r.status = "active" AS is_active
}
WITH profile
WHERE is_active = TRUE
// --------------
// get the attributes that we care about
// --------------
MATCH (profile)-[:HAS_ATTRIBUTE]-(attr:Attribute)
WHERE attr.name IN $attr_names
WITH DISTINCT profile, attr
CALL (profile, attr) {
    MATCH (profile)-[r:HAS_ATTRIBUTE]->(attr)
    WHERE %(branch_filter)s
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    RETURN r.status = "active" AS is_active
}
WITH profile, attr
WHERE is_active = TRUE
// --------------
// get the attribute values
// --------------
CALL (attr) {
    MATCH (attr)-[r:HAS_VALUE]->(av)
    WHERE %(branch_filter)s
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    RETURN av, r.status = "active" AS is_active
    LIMIT 1
}
WITH profile, attr, av
WHERE is_active = TRUE
RETURN profile.uuid AS profile_uuid, attr.name AS attr_name, av.value AS attr_value
        """ % {"branch_filter": branch_filter}
        self.add_to_query(query)
        self.return_labels = ["profile_uuid", "attr_name", "attr_value"]

    def get_profile_data(self) -> list[ProfileData]:
        profile_data_by_uuid: dict[str, ProfileData] = {}
        for result in self.results:
            profile_uuid = result.get_as_type(label="profile_uuid", return_type=str)
            if profile_uuid not in profile_data_by_uuid:
                profile_data_by_uuid[profile_uuid] = ProfileData(
                    uuid=profile_uuid, priority=float("inf"), attribute_values={}
                )
            profile_data = profile_data_by_uuid[profile_uuid]
            attr_name = result.get_as_type(label="attr_name", return_type=str)
            attr_value: Any = result.get(label="attr_value")
            if attr_value == NULL_VALUE:
                attr_value = None
            if attr_name == "profile_priority":
                if attr_value is not None and not isinstance(attr_value, int):
                    attr_value = int(attr_value)
                profile_data.priority = attr_value
            else:
                profile_data.attribute_values[attr_name] = attr_value
        return list(profile_data_by_uuid.values())



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.initialization import get_root_node
from infrahub.log import get_logger

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


log = get_logger()


async def validate_graph_version(db: InfrahubDatabase) -> None:
    root = await get_root_node(db=db)
    if root.graph_version != GRAPH_VERSION:
        log.warning(
            f"Expected database graph version {GRAPH_VERSION} but got {root.graph_version}, possibly 'infrahub upgrade' has not been executed"
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

import jwt
import ujson
from authlib.integrations.httpx_client import AsyncOAuth2Client
from fastapi import APIRouter, Depends, Request, Response
from fastapi.responses import JSONResponse, RedirectResponse
from opentelemetry import trace
from pydantic import BaseModel, HttpUrl

from infrahub import config, models
from infrahub.api.dependencies import get_db
from infrahub.auth import (
    get_groups_from_provider,
    signin_sso_account,
    validate_auth_response,
)
from infrahub.exceptions import ProcessingError
from infrahub.log import get_logger
from infrahub.message_bus.types import KVTTL

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase
    from infrahub.services import InfrahubServices


log = get_logger()
router = APIRouter(prefix="/oidc")


class OIDCDiscoveryConfig(BaseModel):
    issuer: HttpUrl
    authorization_endpoint: HttpUrl
    token_endpoint: HttpUrl
    userinfo_endpoint: HttpUrl
    jwks_uri: HttpUrl
    revocation_endpoint: HttpUrl | None = None
    registration_endpoint: HttpUrl | None = None
    introspection_endpoint: HttpUrl | None = None
    end_session_endpoint: HttpUrl | None = None
    frontchannel_logout_supported: bool | None = None
    frontchannel_logout_session_supported: bool | None = None
    grant_types_supported: list[str] | None = None
    response_types_supported: list[str]
    subject_types_supported: list[str]
    id_token_signing_alg_values_supported: list[str]
    scopes_supported: list[str] | None = None
    token_endpoint_auth_methods_supported: list[str] | None = None
    claims_supported: list[str] | None = None
    acr_values_supported: list[str] | None = None
    request_parameter_supported: bool | None = None
    request_uri_parameter_supported: bool | None = None
    require_request_uri_registration: bool | None = None
    code_challenge_methods_supported: list[str] | None = None
    tls_client_certificate_bound_access_tokens: bool | None = None
    mtls_endpoint_aliases: dict[str, HttpUrl] | None = None


def _get_redirect_url(request: Request, provider_name: str) -> str:
    """Return public redirect URL."""
    base_url = config.SETTINGS.main.public_url or str(request.base_url)
    return urljoin(base_url, f"auth/oidc/{provider_name}/callback")


@router.get("/{provider_name:str}/authorize")
async def authorize(request: Request, provider_name: str, final_url: str | None = None) -> Response:
    provider = config.SETTINGS.security.get_oidc_provider(provider=provider_name)
    service: InfrahubServices = request.app.state.service

    response = await service.http.get(url=provider.discovery_url)
    validate_auth_response(response=response, provider_type="OIDC")
    oidc_config = OIDCDiscoveryConfig(**response.json())

    with trace.get_tracer(__name__).start_as_current_span("sso_oauth2_client_configuration") as span:
        span.set_attribute("provider_name", provider_name)
        span.set_attribute("scopes", provider.scopes)
        span.set_attribute("discovery_url", provider.discovery_url)

        client = AsyncOAuth2Client(
            client_id=provider.client_id,
            client_secret=provider.client_secret,
            scope=provider.scopes,
        )

    redirect_uri = _get_redirect_url(request=request, provider_name=provider_name)
    final_url = final_url or config.SETTINGS.main.public_url or str(request.base_url)

    authorization_uri, state = client.create_authorization_url(
        url=str(oidc_config.authorization_endpoint), redirect_uri=redirect_uri, scope=provider.scopes
    )

    await service.cache.set(
        key=f"security:oidc:provider:{provider_name}:state:{state}", value=final_url, expires=KVTTL.TWO_HOURS
    )

    if config.SETTINGS.dev.frontend_redirect_sso:
        return JSONResponse(content={"url": authorization_uri})

    return RedirectResponse(url=authorization_uri)


@router.get("/{provider_name:str}/token")
async def token(
    request: Request,
    response: Response,
    provider_name: str,
    state: str,
    code: str,
    db: InfrahubDatabase = Depends(get_db),
) -> models.UserTokenWithUrl:
    provider = config.SETTINGS.security.get_oidc_provider(provider=provider_name)

    service: InfrahubServices = request.app.state.service

    cache_key = f"security:oidc:provider:{provider_name}:state:{state}"
    stored_final_url = await service.cache.get(key=cache_key)
    await service.cache.delete(key=cache_key)

    if not stored_final_url:
        raise ProcessingError(message="Invalid 'state' parameter")

    token_data = {
        "code": code,
        "client_id": provider.client_id,
        "client_secret": provider.client_secret,
        "redirect_uri": _get_redirect_url(request=request, provider_name=provider_name),
        "grant_type": "authorization_code",
    }

    discovery_response = await service.http.get(url=provider.discovery_url)
    validate_auth_response(response=discovery_response, provider_type="OIDC")

    oidc_config = OIDCDiscoveryConfig(**discovery_response.json())

    token_response = await service.http.post(str(oidc_config.token_endpoint), data=token_data)
    validate_auth_response(response=token_response, provider_type="OIDC")

    with trace.get_tracer(__name__).start_as_current_span("sso_token_request") as span:
        span.set_attribute("token_request_data", ujson.dumps(token_response.json()))
        payload: dict[str, Any] = token_response.json()

    headers = {"Authorization": f"{payload.get('token_type')} {payload.get('access_token')}"}

    if provider.userinfo_method == config.UserInfoMethod.GET:
        userinfo_response = await service.http.get(str(oidc_config.userinfo_endpoint), headers=headers)
    else:
        userinfo_response = await service.http.post(str(oidc_config.userinfo_endpoint), headers=headers)

    validate_auth_response(response=userinfo_response, provider_type="OIDC")
    user_info: dict[str, Any] = userinfo_response.json()
    sso_groups = (
        user_info.get("groups")
        or await _get_id_token_groups(
            oidc_config=oidc_config, service=service, payload=payload, client_id=provider.client_id
        )
        or await get_groups_from_provider(provider=provider, service=service, payload=payload, user_info=user_info)
    )

    log.info(
        "SSO user authenticated",
        body={"user_name": user_info.get("name"), "groups": sso_groups},
    )

    if not sso_groups and config.SETTINGS.security.sso_user_default_group:
        sso_groups = [config.SETTINGS.security.sso_user_default_group]

    with trace.get_tracer(__name__).start_as_current_span("signin_sso_account") as span:
        span.set_attribute("account_name", ujson.dumps(userinfo_response.json()))
        span.set_attribute("sso_groups", sso_groups)
        user_token = await signin_sso_account(db=db, account_name=user_info["name"], sso_groups=sso_groups)

    response.set_cookie(
        "access_token", user_token.access_token, httponly=True, max_age=config.SETTINGS.security.access_token_lifetime
    )
    response.set_cookie(
        "refresh_token",
        user_token.refresh_token,
        httponly=True,
        max_age=config.SETTINGS.security.refresh_token_lifetime,
    )

    return models.UserTokenWithUrl(
        access_token=user_token.access_token, refresh_token=user_token.refresh_token, final_url=stored_final_url
    )


async def _get_id_token_groups(
    oidc_config: OIDCDiscoveryConfig, service: InfrahubServices, payload: dict[str, Any], client_id: str
) -> list[str]:
    id_token = payload.get("id_token")
    if not id_token:
        return []
    jwks = await service.http.get(url=str(oidc_config.jwks_uri))

    jwk_client = jwt.PyJWKClient(uri=str(oidc_config.jwks_uri), cache_jwk_set=True)
    if jwk_client.jwk_set_cache:
        jwk_client.jwk_set_cache.put(jwks.json())

    signing_key = jwk_client.get_signing_key_from_jwt(id_token)

    decoded_token: dict[str, Any] = jwt.decode(
        jwt=id_token,
        key=signing_key.key,
        algorithms=oidc_config.id_token_signing_alg_values_supported,
        audience=client_id,
        issuer=str(oidc_config.issuer),
        options={"verify_signature": False, "verify_aud": False, "verify_iss": False},
    )

    return decoded_token.get("groups", [])



from graphene import InputObjectType, String


class ContextAccountInput(InputObjectType):
    id = String(required=True, description="The Infrahub ID of the account")


class ContextInput(InputObjectType):
    account = ContextAccountInput(
        required=False,
        description="The account context can be used to override the account information that will be associated with the mutation",
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core import registry
from infrahub.core.node import Node
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.core.schema.node_schema import NodeSchema
from infrahub.exceptions import PoolExhaustedError
from infrahub.tasks.registry import update_branch_registry

from ..query import AttributeMigrationQuery, MigrationBaseQuery
from ..query.attribute_add import AttributeAddQuery
from ..shared import AttributeSchemaMigration, MigrationResult

if TYPE_CHECKING:
    from infrahub.core.node.resource_manager.number_pool import CoreNumberPool
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.schema.attribute_schema import AttributeSchema
    from infrahub.database import InfrahubDatabase

    from ...branch import Branch
    from ...timestamp import Timestamp


class NodeAttributeAddMigrationQuery01(AttributeMigrationQuery, AttributeAddQuery):
    name = "migration_node_attribute_add_01"

    def _get_node_kinds(self, schema: MainSchemaTypes, new_attribute_schema: AttributeSchema) -> list[str]:
        schema_kinds = [schema.kind]
        if not isinstance(schema, (NodeSchema, GenericSchema)):
            return schema_kinds
        if new_attribute_schema.support_profiles:
            schema_kinds.append(f"Profile{schema.kind}")
            if isinstance(schema, GenericSchema) and schema.used_by:
                schema_kinds.extend([f"Profile{kind}" for kind in schema.used_by])
        return schema_kinds

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ):
        node_kinds = self._get_node_kinds(
            schema=migration.new_schema, new_attribute_schema=migration.new_attribute_schema
        )
        super().__init__(
            migration=migration,
            node_kinds=node_kinds,
            attribute_name=migration.new_attribute_schema.name,
            attribute_kind=migration.new_attribute_schema.kind,
            branch_support=migration.new_attribute_schema.get_branch().value,
            default_value=migration.new_attribute_schema.default_value,
            **kwargs,
        )


class NodeAttributeAddMigration(AttributeSchemaMigration):
    name: str = "node.attribute.add"
    queries: Sequence[type[AttributeMigrationQuery]] = [NodeAttributeAddMigrationQuery01]  # type: ignore[assignment]

    async def execute(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | str | None = None,
        queries: Sequence[type[MigrationBaseQuery]] | None = None,
    ) -> MigrationResult:
        if self.new_attribute_schema.inherited is True:
            return MigrationResult()
        return await super().execute(db=db, branch=branch, at=at, queries=queries)

    async def execute_post_queries(
        self,
        db: InfrahubDatabase,
        result: MigrationResult,
        branch: Branch,
        at: Timestamp,  # noqa: ARG002
    ) -> MigrationResult:
        if self.new_attribute_schema.kind != "NumberPool":
            return result

        number_pool: CoreNumberPool = await Node.fetch_or_create_number_pool(
            db=db,
            branch=branch,
            schema_node=self.new_schema,  # type: ignore
            schema_attribute=self.new_attribute_schema,
        )

        await update_branch_registry(db=db, branch=branch)

        nodes: list[Node] = await registry.manager.query(
            db=db, branch=branch, schema=self.new_schema, fields={"id": True, self.new_attribute_schema.name: True}
        )

        try:
            numbers = await number_pool.get_next_many(
                db=db,
                branch=branch,
                quantity=len(nodes),
                attribute=self.new_attribute_schema,
            )
        except PoolExhaustedError as exc:
            result.errors.append(str(exc))
            return result

        for node, number in zip(nodes, numbers, strict=True):
            await number_pool.reserve(db=db, number=number, identifier=node.get_id())
            attr = getattr(node, self.new_attribute_schema.name)
            attr.value = number
            attr.source = number_pool.id

            await node.save(db=db, fields=[self.new_attribute_schema.name])

        return result
